@@ -17,19 +17,22 @@ const AppFavorites = imports.ui.appFavorites;
 const DND = imports.ui.dnd;
 const GenericDisplay = imports.ui.genericDisplay;
 const Main = imports.ui.main;
+const Overview = imports.ui.overview;
+const PopupMenu = imports.ui.popupMenu;
 const Search = imports.ui.search;
+const Tweener = imports.ui.tweener;
 const Workspace = imports.ui.workspace;
 
 const APPICON_SIZE = 48;
 const WELL_MAX_COLUMNS = 8;
 const MENU_POPUP_TIMEOUT = 600;
 
-function AllAppView() {
+function AlphabeticalView() {
     this._init();
 }
 
-AllAppView.prototype = {
-    _init: function(apps) {
+AlphabeticalView.prototype = {
+    _init: function() {
         this.actor = new St.BoxLayout({ vertical: true });
         this._grid = new WellGrid(true);
         this._appSystem = Shell.AppSystem.get_default();
@@ -42,17 +45,17 @@ AllAppView.prototype = {
     },
 
     _addApp: function(app) {
-        let App = new AppWellIcon(this._appSystem.get_app(app.get_id()));
-        App.connect('launching', Lang.bind(this, function() {
+        let appIcon = new AppWellIcon(this._appSystem.get_app(app.get_id()));
+        appIcon.connect('launching', Lang.bind(this, function() {
             this.emit('launching');
         }));
-        App._draggable.connect('drag-begin', Lang.bind(this, function() {
+        appIcon._draggable.connect('drag-begin', Lang.bind(this, function() {
             this.emit('drag-begin');
         }));
 
-        this._grid.addItem(App.actor);
+        this._grid.addItem(appIcon.actor);
 
-        this._apps.push(App);
+        this._apps.push(appIcon);
     },
 
     refresh: function(apps) {
@@ -71,7 +74,67 @@ AllAppView.prototype = {
     }
 };
 
-Signals.addSignalMethods(AllAppView.prototype);
+Signals.addSignalMethods(AlphabeticalView.prototype);
+
+function ViewByCategories() {
+    this._init();
+}
+
+ViewByCategories.prototype = {
+    _init: function() {
+        this._appSystem = Shell.AppSystem.get_default();
+        this.actor = new St.BoxLayout({ vertical: true });
+        this.actor._delegate = this;
+        this._sections = [];
+    },
+
+    _updateSections: function(apps) {
+        this._removeAll();
+
+        let sections = this._appSystem.get_sections();
+        if (!sections)
+            return;
+        for (let i = 0; i < sections.length; i++) {
+            if (i) {
+                let actor = new St.Bin({ style_class: 'app-section-divider' });
+                let divider = new St.Bin({ style_class: 'app-section-divider-container',
+                                           child: actor,
+                                           x_fill: true });
+
+                this.actor.add(divider, { y_fill: false, expand: true });
+            }
+            let _apps = apps.filter(function(app) {
+                return app.get_section() == sections[i];
+            });
+            this._sections[i] = { view: new AlphabeticalView(),
+                                  apps: _apps,
+                                  name: sections[i] };
+            this._sections[i].view.connect('launching', Lang.bind(this, function() {
+                this.emit('launching');
+            }));
+            this._sections[i].view.connect('drag-begin', Lang.bind(this, function() {
+                this.emit('drag-begin');
+            }));
+            this.actor.add(this._sections[i].view.actor, { y_align: St.Align.START, expand: true });
+        }
+    },
+
+    _removeAll: function() {
+        this.actor.destroy_children();
+        this._sections.forEach(function (section) { section.view.disconnectAll(); });
+
+        this._sections = [];
+    },
+
+    refresh: function(apps) {
+        this._updateSections(apps);
+        for (let i = 0; i < this._sections.length; i++) {
+            this._sections[i].view.refresh(this._sections[i].apps);
+        }
+    }
+};
+
+Signals.addSignalMethods(ViewByCategories.prototype);
 
 /* This class represents a display containing a collection of application items.
  * The applications are sorted based on their name.
@@ -100,7 +163,7 @@ AllAppDisplay.prototype = {
         this.actor.add(bin);
         this.actor.add(view, { expand: true, y_fill: false, y_align: St.Align.START });
 
-        this._appView = new AllAppView();
+        this._appView = new ViewByCategories();
         this._appView.connect('launching', Lang.bind(this, this.close));
         this._appView.connect('drag-begin', Lang.bind(this, this.close));
         this._scrollView.add_actor(this._appView.actor);
@@ -119,9 +182,28 @@ AllAppDisplay.prototype = {
     },
 
     toggle: function() {
-        this.emit('open-state-changed', !this.actor.visible);
-
-        this.actor.visible = !this.actor.visible;
+        if (this.actor.visible) {
+            Tweener.addTween(this.actor,
+                             { opacity: 0,
+                               time: Overview.PANE_FADE_TIME,
+                               transition: 'easeOutQuad',
+                               onComplete: Lang.bind(this,
+                                   function() {
+                                       this.actor.hide();
+                                       this.emit('open-state-changed',
+                                                 this.actor.visible);
+                                   })
+                             });
+        } else {
+            this.actor.show();
+            this.emit('open-state-changed', this.actor.visible);
+            this.actor.opacity = 0;
+            Tweener.addTween(this.actor,
+                             { opacity: 255,
+                               time: Overview.PANE_FADE_TIME,
+                               transition: 'easeOutQuad'
+                             });
+        }
     },
 
     close: function() {
@@ -374,7 +456,6 @@ AppIcon.prototype = {
         box.add(this.icon, { expand: true, x_fill: false, y_fill: false });
 
         this._name = new St.Label({ text: this.app.get_name() });
-        this._name.clutter_text.line_alignment = Pango.Alignment.CENTER;
         box.add_actor(this._name);
     }
 };
@@ -396,33 +477,35 @@ AppWellIcon.prototype = {
         this.actor.set_child(this._icon.actor);
 
         this.actor.connect('clicked', Lang.bind(this, this._onClicked));
+
         this._menu = null;
+        this._menuManager = new PopupMenu.PopupMenuManager(this);
 
         this._draggable = DND.makeDraggable(this.actor);
         this._draggable.connect('drag-begin', Lang.bind(this,
             function () {
                 this._removeMenuTimeout();
+                Main.overview.beginItemDrag(this);
+            }));
+        this._draggable.connect('drag-end', Lang.bind(this,
+            function () {
+               Main.overview.endItemDrag(this);
             }));
 
         this.actor.connect('button-press-event', Lang.bind(this, this._onButtonPress));
-        this.actor.connect('show', Lang.bind(this, this._onShow));
-        this.actor.connect('hide', Lang.bind(this, this._onHideDestroy));
-        this.actor.connect('destroy', Lang.bind(this, this._onHideDestroy));
+        this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
 
-        this._stateChangedId = 0;
         this._menuTimeoutId = 0;
-    },
-
-    _onShow: function() {
         this._stateChangedId = this.app.connect('notify::state',
                                                 Lang.bind(this,
                                                           this._onStateChanged));
         this._onStateChanged();
     },
 
-    _onHideDestroy: function() {
+    _onDestroy: function() {
         if (this._stateChangedId > 0)
             this.app.disconnect(this._stateChangedId);
+        this._stateChangedId = 0;
         this._removeMenuTimeout();
     },
 
@@ -446,7 +529,7 @@ AppWellIcon.prototype = {
             this._removeMenuTimeout();
             this._menuTimeoutId = Mainloop.timeout_add(MENU_POPUP_TIMEOUT,
                 Lang.bind(this, function() {
-                    this.popupMenu(button);
+                    this.popupMenu();
                 }));
         }
     },
@@ -466,9 +549,7 @@ AppWellIcon.prototype = {
                 Main.overview.hide();
             }
         } else if (button == 3) {
-            // Don't bind to the right click here; we want left click outside the
-            // area to deactivate as well.
-            this.popupMenu(0);
+            this.popupMenu();
         }
         return false;
     },
@@ -477,7 +558,7 @@ AppWellIcon.prototype = {
         return this.app.get_id();
     },
 
-    popupMenu: function(activatingButton) {
+    popupMenu: function() {
         this._removeMenuTimeout();
         this.actor.fake_release();
 
@@ -496,9 +577,12 @@ AppWellIcon.prototype = {
                     this._onMenuPoppedDown();
                 }
             }));
+
+            this._menuManager.addMenu(this._menu, true);
         }
 
-        this._menu.popup(activatingButton);
+        this._menu.popup();
+        this._menuManager.grab();
 
         return false;
     },
@@ -565,6 +649,11 @@ AppWellIcon.prototype = {
         Main.overview.hide();
     },
 
+    // called by this._menuManager when it has the grab
+    menuEventFilter: function(event) {
+        return this._menu.menuEventFilter(event);
+    },
+
     shellWorkspaceLaunch : function() {
         this.app.open_new_window();
     },
@@ -586,103 +675,34 @@ function AppIconMenu(source) {
 }
 
 AppIconMenu.prototype = {
+    __proto__: PopupMenu.PopupMenu.prototype,
+
     _init: function(source) {
+        PopupMenu.PopupMenu.prototype._init.call(this, source.actor, St.Align.MIDDLE, St.Side.LEFT, 0);
+
         this._source = source;
 
-        this.actor = new Shell.GenericContainer({ reactive: true });
-        this.actor.connect('get-preferred-width', Lang.bind(this, this._getPreferredWidth));
-        this.actor.connect('get-preferred-height', Lang.bind(this, this._getPreferredHeight));
-        this.actor.connect('allocate', Lang.bind(this, this._allocate));
+        this.connect('active-changed', Lang.bind(this, this._onActiveChanged));
+        this.connect('activate', Lang.bind(this, this._onActivate));
+        this.connect('open-state-changed', Lang.bind(this, this._onOpenStateChanged));
 
-        this._windowContainer = new Shell.Menu({ style_class: 'app-well-menu',
-                                                 vertical: true,
-                                                 width: Main.overview._dash.actor.width });
-        this._windowContainer.connect('unselected', Lang.bind(this, this._onItemUnselected));
-        this._windowContainer.connect('selected', Lang.bind(this, this._onItemSelected));
-        this._windowContainer.connect('cancelled', Lang.bind(this, this._onWindowSelectionCancelled));
-        this._windowContainer.connect('activate', Lang.bind(this, this._onItemActivate));
-        this.actor.add_actor(this._windowContainer);
-
-        // Stay popped up on release over application icon
-        this._windowContainer.set_persistent_source(this._source.actor);
-
-        // Intercept events while the menu has the pointer grab to do window-related effects
-        this._windowContainer.connect('enter-event', Lang.bind(this, this._onMenuEnter));
-        this._windowContainer.connect('leave-event', Lang.bind(this, this._onMenuLeave));
-        this._windowContainer.connect('button-release-event', Lang.bind(this, this._onMenuButtonRelease));
-
-        this._arrow = new St.DrawingArea({ style_class: 'app-well-menu-arrow' });
-        this._arrow.connect('repaint', Lang.bind(this, function (area) {
-            Shell.draw_box_pointer(area, Shell.PointerDirection.LEFT);
-        }));
-        this.actor.add_actor(this._arrow);
+        this.actor.add_style_class_name('app-well-menu');
 
         // Chain our visibility and lifecycle to that of the source
         source.actor.connect('notify::mapped', Lang.bind(this, function () {
             if (!source.actor.mapped)
-                this._windowContainer.popdown();
+                this.close();
         }));
         source.actor.connect('destroy', Lang.bind(this, function () { this.actor.destroy(); }));
 
         Main.uiGroup.add_actor(this.actor);
     },
 
-    _getPreferredWidth: function(actor, forHeight, alloc) {
-        let [menuMin, menuNatural] = this._windowContainer.get_preferred_width(forHeight);
-        let [arrowMin, arrowNatural] = this._arrow.get_preferred_width(forHeight);
-        alloc.min_size = menuMin + arrowMin;
-        alloc.natural_size = menuNatural + arrowNatural;
-    },
-
-    _getPreferredHeight: function(actor, forWidth, alloc) {
-        let [min, natural] = this._windowContainer.get_preferred_height(forWidth);
-        alloc.min_size = min;
-        alloc.natural_size = natural;
-    },
-
-    _allocate: function(actor, box, flags) {
-        let childBox = new Clutter.ActorBox();
-        let themeNode = this._windowContainer.get_theme_node();
-
-        let width = box.x2 - box.x1;
-        let height = box.y2 - box.y1;
-
-        let [arrowMinWidth, arrowWidth] = this._arrow.get_preferred_width(height);
-
-        childBox.x1 = 0;
-        childBox.x2 = arrowWidth;
-        childBox.y1 = Math.floor((height / 2) - (arrowWidth / 2));
-        childBox.y2 = childBox.y1 + arrowWidth;
-        this._arrow.allocate(childBox, flags);
-
-        // Ensure the arrow is above the border area
-        let border = themeNode.get_border_width(St.Side.LEFT);
-        childBox.x1 = arrowWidth - border;
-        childBox.x2 = width;
-        childBox.y1 = 0;
-        childBox.y2 = height;
-        this._windowContainer.allocate(childBox, flags);
-    },
-
     _redisplay: function() {
-        this._windowContainer.remove_all();
+        this.removeAll();
 
         let windows = this._source.app.get_windows();
 
-        this._windowContainer.show();
-
-        let iconsDiffer = false;
-        let texCache = St.TextureCache.get_default();
-        if (windows.length > 0) {
-            let firstIcon = windows[0].mini_icon;
-            for (let i = 1; i < windows.length; i++) {
-                if (!texCache.pixbuf_equal(windows[i].mini_icon, firstIcon)) {
-                    iconsDiffer = true;
-                    break;
-                }
-            }
-        }
-
         // Display the app windows menu items and the separator between windows
         // of the current desktop and other windows.
         let activeWorkspace = global.screen.get_active_workspace();
@@ -693,8 +713,8 @@ AppIconMenu.prototype = {
                 this._appendSeparator();
                 separatorShown = true;
             }
-            let box = this._appendMenuItem(windows[i].title);
-            box._window = windows[i];
+            let item = this._appendMenuItem(windows[i].title);
+            item._window = windows[i];
         }
 
         if (windows.length > 0)
@@ -713,45 +733,53 @@ AppIconMenu.prototype = {
     },
 
     _appendSeparator: function () {
-        let bin = new St.Bin({ style_class: 'app-well-menu-separator' });
-        this._windowContainer.add_actor(bin);
+        let separator = new PopupMenu.PopupSeparatorMenuItem();
+        this.addMenuItem(separator);
     },
 
     _appendMenuItem: function(labelText) {
-        let box = new St.BoxLayout({ style_class: 'app-well-menu-item',
-                                      reactive: true });
-        let label = new St.Label({ text: labelText });
-        box.add(label);
-        this._windowContainer.add_actor(box);
-        return box;
+        // FIXME: app-well-menu-item style
+        let item = new PopupMenu.PopupMenuItem(labelText);
+        this.addMenuItem(item);
+        return item;
     },
 
     popup: function(activatingButton) {
-        let [stageX, stageY] = this._source.actor.get_transformed_position();
-        let [stageWidth, stageHeight] = this._source.actor.get_transformed_size();
-
         this._redisplay();
-
-        this._windowContainer.popup(activatingButton, global.get_current_time());
-
-        this.emit('popup', true);
-
-        let x, y;
-        x = Math.floor(stageX + stageWidth);
-        y = Math.floor(stageY + (stageHeight / 2) - (this.actor.height / 2));
-
-        this.actor.set_position(x, y);
-        this.actor.show();
+        this.open();
     },
 
-    popdown: function() {
-        this._windowContainer.popdown();
-        this.emit('popup', false);
-        this.actor.hide();
+    _onOpenStateChanged: function (menu, open) {
+        if (open) {
+            this.emit('popup', true);
+        } else {
+            this._updateHighlight(null);
+            this.emit('popup', false);
+        }
     },
 
-    selectWindow: function(metaWindow) {
-        this._selectMenuItemForWindow(metaWindow);
+    // called by this._menuManager when it has the grab
+    menuEventFilter: function(event) {
+        let eventType = event.type();
+
+        // Check if the user is interacting with a window representation
+        // rather than interacting with the menu
+
+        if (eventType == Clutter.EventType.BUTTON_RELEASE) {
+            let metaWindow = this._findMetaWindowForActor(event.get_source());
+            if (metaWindow)
+                this.emit('activate-window', metaWindow);
+        } else if (eventType == Clutter.EventType.ENTER) {
+            let metaWindow = this._findMetaWindowForActor(event.get_source());
+            if (metaWindow)
+                this._selectMenuItemForWindow(metaWindow, true);
+        } else if (eventType == Clutter.EventType.LEAVE) {
+            let metaWindow = this._findMetaWindowForActor(event.get_source());
+            if (metaWindow)
+                this._selectMenuItemForWindow(metaWindow, false);
+        }
+
+        return false;
     },
 
     _findMetaWindowForActor: function (actor) {
@@ -762,64 +790,32 @@ AppIconMenu.prototype = {
         return null;
     },
 
-    // This function is called while the menu has a pointer grab; what we want
-    // to do is see if the mouse was released over a window representation
-    _onMenuButtonRelease: function (actor, event) {
-        let metaWindow = this._findMetaWindowForActor(event.get_source());
-        if (metaWindow) {
-            this.emit('activate-window', metaWindow);
-        }
-    },
-
     _updateHighlight: function (item) {
-        if (this._highlightedItem) {
-            this._highlightedItem.remove_style_pseudo_class('hover');
+        if (this._highlightedItem)
             this.emit('highlight-window', null);
-        }
         this._highlightedItem = item;
         if (this._highlightedItem) {
-            item.add_style_pseudo_class('hover');
             let window = this._highlightedItem._window;
             if (window)
                 this.emit('highlight-window', window);
         }
     },
 
-    _selectMenuItemForWindow: function (metaWindow) {
-        let children = this._windowContainer.get_children();
-        for (let i = 0; i < children.length; i++) {
-            let child = children[i];
-            let menuMetaWindow = child._window;
+    _selectMenuItemForWindow: function (metaWindow, selected) {
+        let items = this.getMenuItems();
+        for (let i = 0; i < items.length; i++) {
+            let item = items[i];
+            let menuMetaWindow = item._window;
             if (menuMetaWindow == metaWindow)
-                this._updateHighlight(child);
+                item.setActive(selected);
         }
     },
 
-    // Called while menu has a pointer grab
-    _onMenuEnter: function (actor, event) {
-        let metaWindow = this._findMetaWindowForActor(event.get_source());
-        if (metaWindow) {
-            this._selectMenuItemForWindow(metaWindow);
-        }
-    },
-
-    // Called while menu has a pointer grab
-    _onMenuLeave: function (actor, event) {
-        let metaWindow = this._findMetaWindowForActor(event.get_source());
-        if (metaWindow) {
-            this._updateHighlight(null);
-        }
-    },
-
-    _onItemUnselected: function (actor, child) {
-        this._updateHighlight(null);
-    },
-
-    _onItemSelected: function (actor, child) {
+    _onActiveChanged: function (menu, child) {
         this._updateHighlight(child);
     },
 
-    _onItemActivate: function (actor, child) {
+    _onActivate: function (actor, child) {
         if (child._window) {
             let metaWindow = child._window;
             this.emit('activate-window', metaWindow);
@@ -834,12 +830,7 @@ AppIconMenu.prototype = {
             else
                 favs.addFavorite(this._source.app.get_id());
         }
-        this.popdown();
-    },
-
-    _onWindowSelectionCancelled: function () {
-        this.emit('highlight-window', null);
-        this.popdown();
+        this.close();
     }
 };
 Signals.addSignalMethods(AppIconMenu.prototype);
@@ -911,7 +902,12 @@ WellGrid.prototype = {
             let childYSpacing = Math.max(0, height - childNaturalHeight) / 2;
 
             let childBox = new Clutter.ActorBox();
-            childBox.x1 = Math.floor(x + childXSpacing);
+            if (St.Widget.get_default_direction() == St.TextDirection.RTL) {
+                let _x = box.x2 - (x + width);
+                childBox.x1 = Math.floor(_x - childXSpacing);
+            } else {
+                childBox.x1 = Math.floor(x + childXSpacing);
+            }
             childBox.y1 = Math.floor(y + childYSpacing);
             childBox.x2 = childBox.x1 + width;
             childBox.y2 = childBox.y1 + height;
@@ -936,7 +932,6 @@ WellGrid.prototype = {
         let nColumns = 0;
         let usedWidth = 0;
         while (nColumns < WELL_MAX_COLUMNS &&
-                nColumns < children.length &&
                 (usedWidth + this._item_size <= forWidth)) {
             usedWidth += this._item_size + this._spacing;
             nColumns += 1;
@@ -976,6 +971,7 @@ function AppWell() {
 
 AppWell.prototype = {
     _init : function() {
+        this._placeholderText = null;
         this._menus = [];
         this._menuDisplays = [];
 
@@ -992,7 +988,7 @@ AppWell.prototype = {
 
         this._appSystem.connect('installed-changed', Lang.bind(this, this._queueRedisplay));
         AppFavorites.getAppFavorites().connect('changed', Lang.bind(this, this._queueRedisplay));
-        this._tracker.connect('app-running-changed', Lang.bind(this, this._queueRedisplay));
+        this._tracker.connect('app-state-changed', Lang.bind(this, this._queueRedisplay));
     },
 
     _appIdListToHash: function(apps) {
@@ -1032,10 +1028,14 @@ AppWell.prototype = {
             let display = new AppWellIcon(app);
             this._grid.addItem(display.actor);
         }
+        if (this._placeholderText) {
+            this._placeholderText.destroy();
+            this._placeholderText = null;
+        }
 
         if (running.length == 0 && nFavorites == 0) {
-            let text = new St.Label({ text: _("Drag here to add favorites")});
-            this._grid.actor.set_child(text);
+            this._placeholderText = new St.Label({ text: _("Drag here to add favorites") });
+            this.actor.add_actor(this._placeholderText);
         }
     },
 
