@@ -1,22 +1,45 @@
-/* -*- mode: js2; js2-basic-offset: 4; tab-width: 4; indent-tabs-mode: nil -*- */
+/* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
 
 const Clutter = imports.gi.Clutter;
 const Gtk = imports.gi.Gtk;
+const St = imports.gi.St;
 const Lang = imports.lang;
 const Signals = imports.signals;
 const Tweener = imports.ui.tweener;
+const Main = imports.ui.main;
 
+const Params = imports.misc.params;
+
+// Time to scale down to maxDragActorSize
+const SCALE_ANIMATION_TIME = 0.25;
+// Time to animate to original position on cancel
 const SNAP_BACK_ANIMATION_TIME = 0.25;
+// Time to animate to original position on success
+const REVERT_ANIMATION_TIME = 0.75;
+
+const DragMotionResult = {
+    NO_DROP:   0,
+    COPY_DROP: 1,
+    MOVE_DROP: 2,
+    CONTINUE:  3
+};
+
+const DragDropResult = {
+    FAILURE:  0,
+    SUCCESS:  1,
+    CONTINUE: 2
+};
 
 let eventHandlerActor = null;
 let currentDraggable = null;
+let dragMonitors = [];
 
 function _getEventHandlerActor() {
     if (!eventHandlerActor) {
         eventHandlerActor = new Clutter.Rectangle();
         eventHandlerActor.width = 0;
         eventHandlerActor.height = 0;
-        global.stage.add_actor(eventHandlerActor);
+        Main.uiGroup.add_actor(eventHandlerActor);
         // We connect to 'event' rather than 'captured-event' because the capturing phase doesn't happen
         // when you've grabbed the pointer.
         eventHandlerActor.connect('event',
@@ -27,31 +50,63 @@ function _getEventHandlerActor() {
     return eventHandlerActor;
 }
 
-function _Draggable(actor, manualMode) {
-    this._init(actor, manualMode);
+function addDragMonitor(monitor) {
+    dragMonitors.push(monitor);
+}
+
+function removeMonitor(monitor) {
+    for (let i = 0; i < dragMonitors.length; i++)
+        if (dragMonitors[i] == monitor) {
+            dragMonitors.splice(i, 1);
+            return;
+        }
+}
+
+function _Draggable(actor, params) {
+    this._init(actor, params);
 }
 
 _Draggable.prototype = {
-    _init : function(actor, manualMode) {
+    _init : function(actor, params) {
+        params = Params.parse(params, { manualMode: false,
+                                        restoreOnSuccess: false,
+                                        dragActorMaxSize: undefined,
+                                        dragActorOpacity: undefined });
+
         this.actor = actor;
-        if (!manualMode)
+        if (!params.manualMode)
             this.actor.connect('button-press-event',
                                Lang.bind(this, this._onButtonPress));
+
+        this.actor.connect('destroy', Lang.bind(this, function() {
+            this.disconnectAll();
+        }));
         this._onEventId = null;
 
+        this._restoreOnSuccess = params.restoreOnSuccess;
+        this._dragActorMaxSize = params.dragActorMaxSize;
+        this._dragActorOpacity = params.dragActorOpacity;
+
         this._buttonDown = false; // The mouse button has been pressed and has not yet been released.
         this._dragInProgress = false; // The drag has been started, and has not been dropped or cancelled yet.
-        this._snapBackInProgress = false; // The drag has been cancelled and the item is in the process of snapping back.
+        this._animationInProgress = false; // The drag is over and the item is in the process of animating to its original position (snapping back or reverting).
     },
 
     _onButtonPress : function (actor, event) {
-        // FIXME: we should make sure it's button 1, but we can't currently
-        // check that from JavaScript
+        if (event.get_button() != 1)
+            return false;
+
         if (Tweener.getTweenCount(actor))
             return false;
 
         this._buttonDown = true;
-        this._grabActor();
+        // special case St.Clickable: grabbing the pointer would mess up the
+        // internal state, so we start the drag manually on hover change
+        if (this.actor instanceof St.Clickable)
+            this.actor.connect('notify::hover',
+                               Lang.bind(this, this._onClickableHoverChanged));
+        else
+            this._grabActor();
 
         let [stageX, stageY] = event.get_coords();
         this._dragStartX = stageX;
@@ -59,7 +114,16 @@ _Draggable.prototype = {
 
         return false;
     },
-    
+
+    _onClickableHoverChanged: function(button) {
+        if (button.hover || !button.held)
+            return;
+
+        button.fake_release();
+        this.startDrag(this._dragStartX, this._dragStartY,
+                       global.get_current_time());
+    },
+
     _grabActor: function() {
         Clutter.grab_pointer(this.actor);
         this._onEventId = this.actor.connect('event',
@@ -91,7 +155,7 @@ _Draggable.prototype = {
             this._buttonDown = false;
             if (this._dragInProgress) {
                 return this._dragActorDropped(event);
-            } else if (this._dragActor != null && !this._snapBackInProgress) {
+            } else if (this._dragActor != null && !this._animationInProgress) {
                 // Drag must have been cancelled with Esc.
                 this._dragComplete();
                 return true;
@@ -140,8 +204,8 @@ _Draggable.prototype = {
             this._ungrabActor();
         this._grabEvents();
 
-        this._dragStartX = stageX;
-        this._dragStartY = stageY;
+        this._dragX = this._dragStartX = stageX;
+        this._dragY = this._dragStartY = stageY;
 
         if (this.actor._delegate && this.actor._delegate.getDragActor) {
             this._dragActor = this.actor._delegate.getDragActor(this._dragStartX, this._dragStartY);
@@ -154,10 +218,9 @@ _Draggable.prototype = {
                 // the dragActor over it. Otherwise, center it
                 // around the pointer
                 let [sourceX, sourceY] = this._dragActorSource.get_transformed_position();
-                let [sourceWidth, sourceHeight] = this._dragActorSource.get_transformed_size();
                 let x, y;
-                if (stageX > sourceX && stageX <= sourceX + sourceWidth &&
-                    stageY > sourceY && stageY <= sourceY + sourceHeight) {
+                if (stageX > sourceX && stageX <= sourceX + this._dragActor.width &&
+                    stageY > sourceY && stageY <= sourceY + this._dragActor.height) {
                     x = sourceX;
                     y = sourceY;
                 } else {
@@ -193,6 +256,45 @@ _Draggable.prototype = {
 
         this._dragActor.reparent(this.actor.get_stage());
         this._dragActor.raise_top();
+
+        this._dragOrigOpacity = this._dragActor.opacity;
+        if (this._dragActorOpacity != undefined)
+            this._dragActor.opacity = this._dragActorOpacity;
+
+        this._snapBackX = this._dragStartX + this._dragOffsetX;
+        this._snapBackY = this._dragStartY + this._dragOffsetY;
+        this._snapBackScale = this._dragActor.scale_x;
+
+        if (this._dragActorMaxSize != undefined) {
+            let [scaledWidth, scaledHeight] = this._dragActor.get_transformed_size();
+            let currentSize = Math.max(scaledWidth, scaledHeight);
+            if (currentSize > this._dragActorMaxSize) {
+                let scale = this._dragActorMaxSize / currentSize;
+                let origScale =  this._dragActor.scale_x;
+                let origDragOffsetX = this._dragOffsetX;
+                let origDragOffsetY = this._dragOffsetY;
+
+                // The position of the actor changes as we scale
+                // around the drag position, but we can't just tween
+                // to the final position because that tween would
+                // fight with updates as the user continues dragging
+                // the mouse; instead we do the position computations in
+                // an onUpdate() function.
+                Tweener.addTween(this._dragActor,
+                                 { scale_x: scale * origScale,
+                                   scale_y: scale * origScale,
+                                   time: SCALE_ANIMATION_TIME,
+                                   transition: 'easeOutQuad',
+                                   onUpdate: function() {
+                                       let currentScale = this._dragActor.scale_x / origScale;
+                                       this._dragOffsetX = currentScale * origDragOffsetX;
+                                       this._dragOffsetY = currentScale * origDragOffsetY;
+                                       this._dragActor.set_position(this._dragX + this._dragOffsetX,
+                                                                    this._dragY + this._dragOffsetY);
+                                   },
+                                   onUpdateScope: this });
+            }
+        }
     },
 
     _maybeStartDrag:  function(event) {
@@ -211,27 +313,54 @@ _Draggable.prototype = {
 
     _updateDragPosition : function (event) {
         let [stageX, stageY] = event.get_coords();
+        this._dragX = stageX;
+        this._dragY = stageY;
 
         // If we are dragging, update the position
         if (this._dragActor) {
             this._dragActor.set_position(stageX + this._dragOffsetX,
                                          stageY + this._dragOffsetY);
+
             // Because we want to find out what other actor is located at the current position of this._dragActor,
             // we have to temporarily hide this._dragActor.
             this._dragActor.hide();
             let target = this._dragActor.get_stage().get_actor_at_pos(Clutter.PickMode.ALL,
-                                                                      stageX + this._dragOffsetX,
-                                                                      stageY + this._dragOffsetY);
+                                                                      stageX, stageY);
             this._dragActor.show();
+
+            // We call observers only once per motion with the innermost
+            // target actor. If necessary, the observer can walk the
+            // parent itself.
+            let dragEvent = {
+                x: stageX,
+                y: stageY,
+                dragActor: this._dragActor,
+                targetActor: target
+            };
+            for (let i = 0; i < dragMonitors.length; i++) {
+                let motionFunc = dragMonitors[i].dragMotion;
+                if (motionFunc)
+                    switch (motionFunc(dragEvent)) {
+                        case DragMotionResult.NO_DROP:
+                        case DragMotionResult.COPY_DROP:
+                        case DragMotionResult.MOVE_DROP:
+                            // TODO: set a special cursor or something ;)
+                            return true;
+                        case DragMotionResult.CONTINUE:
+                            continue;
+                    }
+            }
+
             while (target) {
                 if (target._delegate && target._delegate.handleDragOver) {
-                    let [targX, targY] = target.get_transformed_position();
+                    let [r, targX, targY] = target.transform_stage_point(stageX, stageY);
                     // We currently loop through all parents on drag-over even if one of the children has handled it.
                     // We can check the return value of the function and break the loop if it's true if we don't want
                     // to continue checking the parents.
-                    target._delegate.handleDragOver(this.actor._delegate, this._dragActor,
-                                                    (stageX + this._dragOffsetX - targX) / target.scale_x,
-                                                    (stageY + this._dragOffsetY - targY) / target.scale_y,
+                    target._delegate.handleDragOver(this.actor._delegate,
+                                                    this._dragActor,
+                                                    targX,
+                                                    targY,
                                                     event.get_time());
                 }
                 target = target.get_parent();
@@ -249,17 +378,44 @@ _Draggable.prototype = {
         let target = this._dragActor.get_stage().get_actor_at_pos(Clutter.PickMode.ALL,
                                                                   dropX, dropY);
         this._dragActor.show();
+
+        // We call observers only once per motion with the innermost
+        // target actor. If necessary, the observer can walk the
+        // parent itself.
+        let dropEvent = {
+            dropActor: this._dragActor,
+            targetActor: target,
+            clutterEvent: event
+        };
+        for (let i = 0; i < dragMonitors.length; i++) {
+            let dropFunc = dragMonitors[i].dragDrop;
+            if (dropFunc)
+                switch (dropFunc(dropEvent)) {
+                    case DragDropResult.FAILURE:
+                    case DragDropResult.SUCCESS:
+                        return true;
+                    case DragDropResult.CONTINUE:
+                        continue;
+                }
+        }
+
         while (target) {
             if (target._delegate && target._delegate.acceptDrop) {
-                let [targX, targY] = target.get_transformed_position();
-                if (target._delegate.acceptDrop(this.actor._delegate, this._dragActor,
-                                                (dropX - targX) / target.scale_x,
-                                                (dropY - targY) / target.scale_y,
+                let [r, targX, targY] = target.transform_stage_point(dropX, dropY);
+                if (target._delegate.acceptDrop(this.actor._delegate,
+                                                this._dragActor,
+                                                targX,
+                                                targY,
                                                 event.get_time())) {
                     // If it accepted the drop without taking the actor,
-                    // destroy it.
-                    if (this._dragActor.get_parent() == this._dragActor.get_stage())
-                        this._dragActor.destroy();
+                    // handle it ourselves.
+                    if (this._dragActor.get_parent() == this._dragActor.get_stage()) {
+                        if (this._restoreOnSuccess) {
+                            this._restoreDragActor(event.get_time());
+                            return true;
+                        } else
+                            this._dragActor.destroy();
+                    }
 
                     this._dragInProgress = false;
                     this.emit('drag-end', event.get_time(), true);
@@ -275,31 +431,59 @@ _Draggable.prototype = {
         return true;
     },
 
+    // Get position of the drag actor's source if the source is still around,
+    // or return the original location if the actor itself was being dragged
+    // or the source is no longer around.
+    _getRestoreLocation: function() {
+        let locX = this._snapBackX;
+        let locY = this._snapBackY;
+
+        if (this._dragActorSource && this._dragActorSource.visible)
+            [locX, locY] = this._dragActorSource.get_transformed_position();
+        return [locX, locY];
+    },
+
     _cancelDrag: function(eventTime) {
         this._dragInProgress = false;
-        // Snap back to the actor source if the source is still around, snap back 
-        // to the original location if the actor itself was being dragged or the
-        // source is no longer around.
-        let snapBackX = this._dragStartX + this._dragOffsetX;
-        let snapBackY = this._dragStartY + this._dragOffsetY;
-        if (this._dragActorSource && this._dragActorSource.visible) {
-            [snapBackX, snapBackY] = this._dragActorSource.get_transformed_position();
-        }
+        let [snapBackX, snapBackY] = this._getRestoreLocation();
 
-        this._snapBackInProgress = true;
+        this._animationInProgress = true;
         // No target, so snap back
         Tweener.addTween(this._dragActor,
                          { x: snapBackX,
                            y: snapBackY,
+                           scale_x: this._snapBackScale,
+                           scale_y: this._snapBackScale,
+                           opacity: this._dragOrigOpacity,
                            time: SNAP_BACK_ANIMATION_TIME,
-                           transition: "easeOutQuad",
-                           onComplete: this._onSnapBackComplete,
+                           transition: 'easeOutQuad',
+                           onComplete: this._onAnimationComplete,
                            onCompleteScope: this,
                            onCompleteParams: [this._dragActor, eventTime]
                          });
     },
 
-    _onSnapBackComplete : function (dragActor, eventTime) {
+    _restoreDragActor: function(eventTime) {
+        this._dragInProgress = false;
+        [restoreX, restoreY] = this._getRestoreLocation();
+
+        // fade the actor back in at its original location
+        this._dragActor.set_position(restoreX, restoreY);
+        this._dragActor.set_scale(this._snapBackScale, this._snapBackScale);
+        this._dragActor.opacity = 0;
+
+        this._animationInProgress = true;
+        Tweener.addTween(this._dragActor,
+                         { opacity: this._dragOrigOpacity,
+                           time: REVERT_ANIMATION_TIME,
+                           transition: 'easeOutQuad',
+                           onComplete: this._onAnimationComplete,
+                           onCompleteScope: this,
+                           onCompleteParams: [this._dragActor, eventTime]
+                         });
+    },
+
+    _onAnimationComplete : function (dragActor, eventTime) {
         if (this._dragOrigParent) {
             dragActor.reparent(this._dragOrigParent);
             dragActor.set_scale(this._dragOrigScale, this._dragOrigScale);
@@ -309,7 +493,7 @@ _Draggable.prototype = {
         }
         this.emit('drag-end', eventTime, false);
 
-        this._snapBackInProgress = false;
+        this._animationInProgress = false;
         if (!this._buttonDown)
             this._dragComplete();
     },
@@ -326,10 +510,24 @@ Signals.addSignalMethods(_Draggable.prototype);
 /**
  * makeDraggable:
  * @actor: Source actor
- * @manualMode: If given, do not automatically start drag and drop on click
+ * @params: (optional) Additional parameters
  *
  * Create an object which controls drag and drop for the given actor.
+ *
+ * If %manualMode is %true in @params, do not automatically start
+ * drag and drop on click
+ *
+ * If %dragActorMaxSize is present in @params, the drag actor will
+ * be scaled down to be no larger than that size in pixels.
+ *
+ * If %dragActorOpacity is present in @params, the drag actor will
+ * will be set to have that opacity during the drag.
+ *
+ * Note that when the drag actor is the source actor and the drop
+ * succeeds, the actor scale and opacity aren't reset; if the drop
+ * target wants to reuse the actor, it's up to the drop target to
+ * reset these values.
  */
-function makeDraggable(actor, manualMode) {
-    return new _Draggable(actor, manualMode);
+function makeDraggable(actor, params) {
+    return new _Draggable(actor, params);
 }
