@@ -90,9 +90,9 @@ function init() {
 
     _blockMethod('Clutter.Event.get_state', 'Shell.get_event_state',
                  'gjs\'s handling of Clutter.ModifierType is broken. See bug 597292.');
-    _blockMethod('Gdk.Display.get_pointer', 'global.get_pointer',
+    _blockMethod('Gdk.Display.get_device_state', 'global.get_pointer',
                  'gjs\'s handling of Gdk.ModifierType is broken. See bug 597292.');
-    _blockMethod('Gdk.Window.get_pointer', 'global.get_pointer',
+    _blockMethod('Gdk.Window.get_device_position', 'global.get_pointer',
                  'gjs\'s handling of Gdk.ModifierType is broken. See bug 597292.');
 
     // Now close the back door to prevent extensions from trying to
