@@ -30,18 +30,21 @@ WindowManager.prototype = {
         this._destroying = [];
 
         this._switchData = null;
-        this._shellwm.connect('switch-workspace', Lang.bind(this, this._switchWorkspace));
         this._shellwm.connect('kill-switch-workspace', Lang.bind(this, this._switchWorkspaceDone));
+        this._shellwm.connect('kill-window-effects', Lang.bind(this, function (shellwm, actor) {
+            this._minimizeWindowDone(shellwm, actor);
+            this._maximizeWindowDone(shellwm, actor);
+            this._unmaximizeWindowDone(shellwm, actor);
+            this._mapWindowDone(shellwm, actor);
+            this._destroyWindowDone(shellwm, actor);
+        }));
+
+        this._shellwm.connect('switch-workspace', Lang.bind(this, this._switchWorkspace));
         this._shellwm.connect('minimize', Lang.bind(this, this._minimizeWindow));
-        this._shellwm.connect('kill-minimize', Lang.bind(this, this._minimizeWindowDone));
         this._shellwm.connect('maximize', Lang.bind(this, this._maximizeWindow));
-        this._shellwm.connect('kill-maximize', Lang.bind(this, this._maximizeWindowDone));
         this._shellwm.connect('unmaximize', Lang.bind(this, this._unmaximizeWindow));
-        this._shellwm.connect('kill-unmaximize', Lang.bind(this, this._unmaximizeWindowDone));
         this._shellwm.connect('map', Lang.bind(this, this._mapWindow));
-        this._shellwm.connect('kill-map', Lang.bind(this, this._mapWindowDone));
         this._shellwm.connect('destroy', Lang.bind(this, this._destroyWindow));
-        this._shellwm.connect('kill-destroy', Lang.bind(this, this._destroyWindowDone));
 
         this._workspaceSwitcherPopup = null;
         this.setKeybindingHandler('switch_to_workspace_left', Lang.bind(this, this._showWorkspaceSwitcher));
@@ -103,7 +106,7 @@ WindowManager.prototype = {
                            x: xDest,
                            y: 0,
                            time: WINDOW_ANIMATION_TIME,
-                           transition: "easeOutQuad",
+                           transition: 'easeOutQuad',
                            onComplete: this._minimizeWindowDone,
                            onCompleteScope: this,
                            onCompleteParams: [shellwm, actor],
@@ -152,17 +155,15 @@ WindowManager.prototype = {
             return;
         }
 
-        actor.move_anchor_point_from_gravity(Clutter.Gravity.CENTER);
-        actor.set_scale(0.0, 0.0);
+        actor.opacity = 0;
         actor.show();
-        
-        /* scale window up from 0x0 to normal size */
+
+        /* Fade window in */
         this._mapping.push(actor);
         Tweener.addTween(actor,
-                         { scale_x: 1.0,
-                           scale_y: 1.0,
+                         { opacity: 255,
                            time: WINDOW_ANIMATION_TIME,
-                           transition: "easeOutQuad",
+                           transition: 'easeOutQuad',
                            onComplete: this._mapWindowDone,
                            onCompleteScope: this,
                            onCompleteParams: [shellwm, actor],
@@ -175,8 +176,7 @@ WindowManager.prototype = {
     _mapWindowDone : function(shellwm, actor) {
         if (this._removeEffect(this._mapping, actor)) {
             Tweener.removeTweens(actor);
-            actor.set_scale(1.0, 1.0);
-            actor.move_anchor_point_from_gravity(Clutter.Gravity.NORTH_WEST);
+            actor.opacity = 255;
             shellwm.completed_map(actor);
         }
     },
@@ -200,7 +200,7 @@ WindowManager.prototype = {
             return;
         }
 
-        let windows = shellwm.get_switch_workspace_actors();
+        let windows = global.get_windows();
 
         /* @direction is the direction that the "camera" moves, so the
          * screen contents have to move one screen's worth in the
@@ -261,7 +261,7 @@ WindowManager.prototype = {
                          { x: xDest,
                            y: yDest,
                            time: WINDOW_ANIMATION_TIME,
-                           transition: "easeOutQuad",
+                           transition: 'easeOutQuad',
                            onComplete: this._switchWorkspaceDone,
                            onCompleteScope: this,
                            onCompleteParams: [shellwm]
@@ -270,7 +270,7 @@ WindowManager.prototype = {
                          { x: 0,
                            y: 0,
                            time: WINDOW_ANIMATION_TIME,
-                           transition: "easeOutQuad"
+                           transition: 'easeOutQuad'
                          });
     },
 
@@ -309,7 +309,7 @@ WindowManager.prototype = {
 
     _showWorkspaceSwitcher : function(shellwm, binding, window, backwards) {
         /* We don't support this kind of layout */
-        if (binding == "switch_to_workspace_up" || binding == "switch_to_workspace_down")
+        if (binding == 'switch_to_workspace_up' || binding == 'switch_to_workspace_down')
             return;
 
         if (global.screen.n_workspaces == 1)
@@ -318,11 +318,11 @@ WindowManager.prototype = {
         if (this._workspaceSwitcherPopup == null)
             this._workspaceSwitcherPopup = new WorkspaceSwitcherPopup.WorkspaceSwitcherPopup();
 
-        if (binding == "switch_to_workspace_left") {
+        if (binding == 'switch_to_workspace_left') {
             this.actionMoveWorkspaceLeft();
         }
 
-        if (binding == "switch_to_workspace_right") {
+        if (binding == 'switch_to_workspace_right') {
             this.actionMoveWorkspaceRight();
         }
     },
