@@ -1,6 +1,5 @@
 /* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
 
-const Big = imports.gi.Big;
 const Clutter = imports.gi.Clutter;
 const GdkPixbuf = imports.gi.GdkPixbuf;
 const Gdk = imports.gi.Gdk;
@@ -29,7 +28,8 @@ FRAME_COLOR.from_pixel(0xffffffff);
 
 const SCROLL_SCALE_AMOUNT = 100 / 5;
 
-const ZOOM_OVERLAY_FADE_TIME = 0.15;
+const LIGHTBOX_FADE_TIME = 0.1;
+const CLOSE_BUTTON_FADE_TIME = 0.1;
 
 const DRAGGING_WINDOW_OPACITY = 100;
 
@@ -122,19 +122,18 @@ WindowClone.prototype = {
                            Lang.bind(this, this._onScroll));
 
         this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
-        this.actor.connect('enter-event',
-                           Lang.bind(this, this._onEnter));
         this.actor.connect('leave-event',
                            Lang.bind(this, this._onLeave));
-        this._havePointer = false;
 
         this._draggable = DND.makeDraggable(this.actor,
-                                            { dragActorMaxSize: WINDOW_DND_SIZE,
+                                            { restoreOnSuccess: true,
+                                              dragActorMaxSize: WINDOW_DND_SIZE,
                                               dragActorOpacity: DRAGGING_WINDOW_OPACITY });
         this._draggable.connect('drag-begin', Lang.bind(this, this._onDragBegin));
         this._draggable.connect('drag-end', Lang.bind(this, this._onDragEnd));
         this.inDrag = false;
 
+        this._windowIsZooming = false;
         this._zooming = false;
         this._selected = false;
     },
@@ -179,23 +178,7 @@ WindowClone.prototype = {
         this.disconnectAll();
     },
 
-    _onEnter: function (actor, event) {
-        // If the user drags faster than we can follow, he'll end up
-        // leaving the window temporarily and then re-entering it
-        if (this.inDrag)
-            return;
-
-        this._havePointer = true;
-    },
-
     _onLeave: function (actor, event) {
-        // If the user drags faster than we can follow, he'll end up
-        // leaving the window temporarily and then re-entering it
-        if (this.inDrag)
-            return;
-
-        this._havePointer = false;
-
         if (this._zoomStep)
             this._zoomEnd();
     },
@@ -235,7 +218,10 @@ WindowClone.prototype = {
         this._zooming = true;
         this.emit('zoom-start');
 
-        this._zoomLightbox = new Lightbox.Lightbox(global.stage, false);
+        if (!this._zoomLightbox)
+            this._zoomLightbox = new Lightbox.Lightbox(global.stage,
+                                                       { fadeTime: LIGHTBOX_FADE_TIME });
+        this._zoomLightbox.show();
 
         this._zoomLocalOrig  = new ScaledPoint(this.actor.x, this.actor.y, this.actor.scale_x, this.actor.scale_y);
         this._zoomGlobalOrig = new ScaledPoint();
@@ -274,7 +260,7 @@ WindowClone.prototype = {
         [this.actor.x, this.actor.y]             = this._zoomLocalOrig.getPosition();
         [this.actor.scale_x, this.actor.scale_y] = this._zoomLocalOrig.getScale();
 
-        this._zoomLightbox.destroy();
+        this._zoomLightbox.hide();
 
         this._zoomLocalPosition  = undefined;
         this._zoomLocalScale     = undefined;
@@ -282,7 +268,6 @@ WindowClone.prototype = {
         this._zoomGlobalScale    = undefined;
         this._zoomTargetPosition = undefined;
         this._zoomStep           = undefined;
-        this._zoomLightbox       = undefined;
     },
 
     _onButtonRelease : function (actor, event) {
@@ -298,14 +283,6 @@ WindowClone.prototype = {
     _onDragEnd : function (draggable, time, snapback) {
         this.inDrag = false;
 
-        // Most likely, the clone is going to move away from the
-        // pointer now. But that won't cause a leave-event, so
-        // do this by hand. Of course, if the window only snaps
-        // back a short distance, this might be wrong, but it's
-        // better to have the label mysteriously missing than
-        // mysteriously present
-        this._havePointer = false;
-
         // We may not have a parent if DnD completed successfully, in
         // which case our clone will shortly be destroyed and replaced
         // with a new one on the target workspace.
@@ -352,7 +329,7 @@ DesktopClone.prototype = {
             Tweener.addTween(this._desktop,
                              { opacity: 255,
                                time: Overview.ANIMATION_TIME,
-                               transition: "easeOutQuad" });
+                               transition: 'easeOutQuad' });
         }
     },
 
@@ -366,7 +343,7 @@ DesktopClone.prototype = {
             Tweener.addTween(this._desktop,
                              { opacity: 0,
                                time: Overview.ANIMATION_TIME,
-                               transition: "easeOutQuad",
+                               transition: 'easeOutQuad',
                                onComplete: Lang.bind(this,
                                    function() {
                                        this._desktop.hide();
@@ -401,7 +378,7 @@ WindowOverlay.prototype = {
         this._windowClone = windowClone;
         this._parentActor = parentActor;
 
-        let title = new St.Label({ style_class: "window-caption",
+        let title = new St.Label({ style_class: 'window-caption',
                                    text: metaWindow.title });
         title.clutter_text.ellipsize = Pango.EllipsizeMode.END;
         title._spacing = 0;
@@ -411,7 +388,7 @@ WindowOverlay.prototype = {
                 this.title.text = w.title;
             }));
 
-        let button = new St.Button({ style_class: "window-close" });
+        let button = new St.Button({ style_class: 'window-close' });
         button._overlap = 0;
 
         this._idleToggleCloseId = 0;
@@ -450,7 +427,7 @@ WindowOverlay.prototype = {
     },
 
     show: function() {
-        let [child, x, y, mask] = Gdk.Screen.get_default().get_root_window().get_pointer();
+        let [x, y, mask] = global.get_pointer();
         let actor = global.stage.get_actor_at_pos(Clutter.PickMode.REACTIVE,
                                                   x, y);
         if (actor == this._windowClone.actor) {
@@ -465,8 +442,8 @@ WindowOverlay.prototype = {
         this._parentActor.raise_top();
         Tweener.addTween(this.title,
                         { opacity: 255,
-                          time: Overview.ANIMATION_TIME,
-                          transition: "easeOutQuad" });
+                          time: CLOSE_BUTTON_FADE_TIME,
+                          transition: 'easeOutQuad' });
     },
 
     chromeWidth: function () {
@@ -560,7 +537,7 @@ WindowOverlay.prototype = {
 
     _idleToggleCloseButton: function() {
         this._idleToggleCloseId = 0;
-        let [child, x, y, mask] = Gdk.Screen.get_default().get_root_window().get_pointer();
+        let [x, y, mask] = global.get_pointer();
         let actor = global.stage.get_actor_at_pos(Clutter.PickMode.REACTIVE,
                                                   x, y);
         if (actor != this._windowClone.actor && actor != this.closeButton) {
@@ -652,10 +629,11 @@ Workspace.prototype = {
         this._desktop.connect('selected',
                               Lang.bind(this,
                                         function(clone, time) {
-                                            // Only switch to the workspace when one window is open
-                                            // We check for "2" here because the Desktop does not really count
-                                            // as a window in this context.
-                                            if (this._windows.length <= 2) {
+                                            // Only switch to the workspace when there's no application windows
+                                            // open (we always have one window for the desktop).  The problem
+                                            // is that it's too easy to miss an app window and get the wrong
+                                            // one focused.
+                                            if (this._windows.length == 1) {
                                                 this.metaWorkspace.activate(time);
                                                 Main.overview.hide();
                                             }
@@ -739,15 +717,17 @@ Workspace.prototype = {
      * This function also resets the highlighted window state.
      */
     setLightboxMode: function (showLightbox) {
-        if (showLightbox) {
-            this._lightbox = new Lightbox.Lightbox(this.actor, false);
-        } else {
-            this._lightbox.destroy();
-            this._lightbox = null;
-        }
-        if (this._frame) {
+        if (!this._lightbox)
+            this._lightbox = new Lightbox.Lightbox(this.actor,
+                                                   { fadeTime: LIGHTBOX_FADE_TIME });
+
+        if (showLightbox)
+            this._lightbox.show();
+        else
+            this._lightbox.hide();
+
+        if (this._frame)
             this._frame.set_opacity(showLightbox ? 150 : 255);
-        }
     },
 
     /**
@@ -757,6 +737,9 @@ Workspace.prototype = {
      * Draw the user's attention to the given window @metaWindow.
      */
     setHighlightWindow: function (metaWindow) {
+        if (!this._lightbox)
+            return;
+
         let actor;
         if (metaWindow != null) {
             let clone = this.lookupCloneForMetaWindow(metaWindow);
@@ -1049,8 +1032,13 @@ Workspace.prototype = {
                              (desiredHeight - buttonOuterHeight - captionHeight) / rect.height,
                              1.0 / this.scale);
 
-        x = xCenter - 0.5 * scale * rect.width;
-        y = y + height - rect.height * scale - captionHeight;
+        x = Math.floor(xCenter - 0.5 * scale * rect.width);
+
+        // We want to center the window in case we have just one
+        if (metaWindow.get_workspace().n_windows == 1)
+            y = Math.floor(yCenter * global.screen_height - 0.5 * scale * rect.height);
+        else
+            y = Math.floor(y + height - rect.height * scale - captionHeight);
 
         return [x, y, scale];
     },
@@ -1126,7 +1114,7 @@ Workspace.prototype = {
                      Tweener.addTween(clone.actor,
                                       { opacity: 255,
                                         time: Overview.ANIMATION_TIME,
-                                        transition: "easeInQuad"
+                                        transition: 'easeInQuad'
                                       });
                 }
 
@@ -1137,7 +1125,7 @@ Workspace.prototype = {
                                    scale_y: scale,
                                    workspace_relative: workspaceZooming ? this : null,
                                    time: Overview.ANIMATION_TIME,
-                                   transition: "easeOutQuad",
+                                   transition: 'easeOutQuad',
                                    onComplete: Lang.bind(this, function() {
                                       this._fadeInWindowOverlay(clone, overlay);
                                    })
@@ -1169,6 +1157,9 @@ Workspace.prototype = {
     },
 
     _fadeInWindowOverlay: function(clone, overlay) {
+        if (clone.inDrag)
+            return;
+
         // This is a little messy and complicated because when we
         // start the fade-in we may not have done the final positioning
         // of the workspaces. (Tweener doesn't necessarily finish
@@ -1209,8 +1200,10 @@ Workspace.prototype = {
     },
 
     _delayedWindowRepositioning: function() {
-        let [child, x, y, mask] =
-            Gdk.Screen.get_default().get_root_window().get_pointer();
+        if (this._windowIsZooming)
+            return true;
+
+        let [x, y, mask] = global.get_pointer();
         let wsWidth = this.actor.width * this.scale;
         let wsHeight = this.actor.height * this.scale;
 
@@ -1282,8 +1275,7 @@ Workspace.prototype = {
         }
 
         // setup new handler
-        let [child, x, y, mask] =
-            Gdk.Screen.get_default().get_root_window().get_pointer();
+        let [x, y, mask] = global.get_pointer();
         this._cursorX = x;
         this._cursorY = y;
 
@@ -1387,7 +1379,7 @@ Workspace.prototype = {
                                    workspace_relative: this,
                                    time: Overview.ANIMATION_TIME,
                                    opacity: 255,
-                                   transition: "easeOutQuad"
+                                   transition: 'easeOutQuad'
                                  });
             } else {
                 // The window is hidden, make it shrink and fade it out
@@ -1397,7 +1389,7 @@ Workspace.prototype = {
                                    opacity: 0,
                                    workspace_relative: this,
                                    time: Overview.ANIMATION_TIME,
-                                   transition: "easeOutQuad"
+                                   transition: 'easeOutQuad'
                                  });
             }
         }
@@ -1420,7 +1412,7 @@ Workspace.prototype = {
                            scale_x: this.scale,
                            scale_y: this.scale,
                            time: Overview.ANIMATION_TIME,
-                           transition: "easeOutQuad",
+                           transition: 'easeOutQuad',
                            onComplete: Lang.bind(this, this._fadeInAllOverlays)
                          });
     },
@@ -1440,7 +1432,7 @@ Workspace.prototype = {
                            scale_x: this.scale,
                            scale_y: this.scale,
                            time: Overview.ANIMATION_TIME,
-                           transition: "easeOutQuad"
+                           transition: 'easeOutQuad'
                          });
 
         this._visible = true;
@@ -1462,7 +1454,7 @@ Workspace.prototype = {
                            scale_x: this.scale,
                            scale_y: this.scale,
                            time: Overview.ANIMATION_TIME,
-                           transition: "easeOutQuad",
+                           transition: 'easeOutQuad',
                            onComplete: onComplete
                          });
 
@@ -1533,6 +1525,14 @@ Workspace.prototype = {
                           this.emit('window-drag-end', clone.actor);
                           overlay.show();
                       }));
+        clone.connect('zoom-start',
+                      Lang.bind(this, function() {
+                          this._windowIsZooming = true;
+                      }));
+        clone.connect('zoom-end',
+                      Lang.bind(this, function() {
+                          this._windowIsZooming = false;
+                      }));
 
         this.actor.add_actor(clone.actor);
 
