@@ -26,13 +26,13 @@ const WORKSPACE_SWITCH_TIME = 0.25;
 // Note that mutter has a compile-time limit of 36
 const MAX_WORKSPACES = 16;
 
-// The values here are also used for gconf, and the key and value
+// The values here are also used for GSettings, and the key and value
 // names must match
 const WorkspacesViewType = {
     SINGLE: 'single',
     GRID:   'grid'
 };
-const WORKSPACES_VIEW_KEY = 'overview/workspaces_view';
+const WORKSPACES_VIEW_KEY = 'workspaces-view';
 
 const WORKSPACE_DRAGGING_SCALE = 0.85;
 const WORKSPACE_SHADOW_SCALE = (1 - WORKSPACE_DRAGGING_SCALE) / 2;
@@ -43,12 +43,10 @@ function GenericWorkspacesView(width, height, x, y, workspaces) {
 
 GenericWorkspacesView.prototype = {
     _init: function(width, height, x, y, workspaces) {
-        this.actor = new St.Bin({ style_class: 'workspaces' });
-        this._actor = new Clutter.Group();
+        this.actor = new St.Group({ style_class: 'workspaces' });
 
         this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
 
-        this.actor.add_actor(this._actor);
         this.actor.connect('style-changed', Lang.bind(this,
             function() {
                 let node = this.actor.get_theme_node();
@@ -74,7 +72,7 @@ GenericWorkspacesView.prototype = {
 
         // Add workspace actors
         for (let w = 0; w < global.screen.n_workspaces; w++)
-            this._workspaces[w].actor.reparent(this._actor);
+            this._workspaces[w].actor.reparent(this.actor);
         this._workspaces[activeWorkspaceIndex].actor.raise_top();
 
         // Position/scale the desktop windows and their children after the
@@ -297,10 +295,10 @@ MosaicView.prototype = {
         GenericWorkspacesView.prototype._init.call(this, width, height, x, y, workspaces);
 
         this.actor.add_style_class_name('mosaic');
-        this._actor.set_clip(x - Workspace.FRAME_SIZE,
-                             y - Workspace.FRAME_SIZE,
-                             width + 2 * Workspace.FRAME_SIZE,
-                             height + 2 * Workspace.FRAME_SIZE);
+        this.actor.set_clip(x - Workspace.FRAME_SIZE,
+                            y - Workspace.FRAME_SIZE,
+                            width + 2 * Workspace.FRAME_SIZE,
+                            height + 2 * Workspace.FRAME_SIZE);
         this._workspaces[global.screen.get_active_workspace_index()].setSelected(true);
     },
 
@@ -411,7 +409,7 @@ MosaicView.prototype = {
         // Add actors
         if (newNumWorkspaces > oldNumWorkspaces)
             for (let w = oldNumWorkspaces; w < newNumWorkspaces; w++)
-                this._actor.add_actor(this._workspaces[w].actor);
+                this.actor.add_actor(this._workspaces[w].actor);
 
         // Figure out the new layout
         this._computeWorkspacePositions();
@@ -629,31 +627,64 @@ SingleView.prototype = {
     __proto__: GenericWorkspacesView.prototype,
 
     _init: function(width, height, x, y, workspaces) {
+        let shadowWidth = Math.ceil(global.screen_width * WORKSPACE_SHADOW_SCALE);
+
         this._newWorkspaceArea = new NewWorkspaceArea();
+        this._newWorkspaceArea.actor._delegate = {
+            acceptDrop: Lang.bind(this, this._acceptNewWorkspaceDrop)
+        };
         this._leftShadow = new St.Bin({ style_class: 'left-workspaces-shadow',
-                                        width: Math.ceil(global.screen_width * WORKSPACE_SHADOW_SCALE),
+                                        width: shadowWidth,
                                         height: global.screen_height,
                                         x: global.screen_width });
+        this._leftShadow._delegate = {
+            acceptDrop: Lang.bind(this, function(source, actor, x, y, time) {
+                let active = global.screen.get_active_workspace_index();
+                let leftWorkspace = this._workspaces[active - 1];
+                if (leftWorkspace &&
+                    leftWorkspace.acceptDrop(source, actor, x, y, time)) {
+                    leftWorkspace.metaWorkspace.activate(time);
+                    return true;
+                }
+                return false;
+            })
+        };
         this._rightShadow = new St.Bin({ style_class: 'right-workspaces-shadow',
-                                         width: Math.ceil(global.screen_width * WORKSPACE_SHADOW_SCALE),
+                                         width: shadowWidth,
                                          height: global.screen_height,
                                          x: global.screen_width });
+        this._rightShadow._delegate = {
+            acceptDrop: Lang.bind(this, function(source, actor, x, y, time) {
+                let active = global.screen.get_active_workspace_index();
+                let rightWorkspace = this._workspaces[active + 1];
+                if (rightWorkspace &&
+                    rightWorkspace.acceptDrop(source, actor, x, y, time)) {
+                    rightWorkspace.metaWorkspace.activate(time);
+                    return true;
+                }
+                return false;
+            })
+        };
 
         GenericWorkspacesView.prototype._init.call(this, width, height, x, y, workspaces);
 
+        this._itemDragBeginId = Main.overview.connect('item-drag-begin',
+                                                      Lang.bind(this, this._dragBegin));
+        this._itemDragEndId = Main.overview.connect('item-drag-end',
+                                                     Lang.bind(this, this._dragEnd));
         for (let i = 0; i < this._workspaces.length; i++) {
             this._workspaces[i]._windowDragBeginId = this._workspaces[i].connect('window-drag-begin',
-                                                                                 Lang.bind(this, this._onWindowDragBegin));
+                                                                                 Lang.bind(this, this._dragBegin));
             this._workspaces[i]._windowDragEndId = this._workspaces[i].connect('window-drag-end',
-                                                                               Lang.bind(this, this._onWindowDragEnd));
+                                                                               Lang.bind(this, this._dragEnd));
         }
 
-        this._actor.add_actor(this._newWorkspaceArea.actor);
-        this._actor.add_actor(this._leftShadow);
-        this._actor.add_actor(this._rightShadow);
+        this.actor.add_actor(this._newWorkspaceArea.actor);
+        this.actor.add_actor(this._leftShadow);
+        this.actor.add_actor(this._rightShadow);
 
         this.actor.add_style_class_name('single');
-        this._actor.set_clip(x, y, width, height);
+        this.actor.set_clip(x, y, width, height);
         this._activeWorkspaceX = 0; // x offset of active ws while dragging
         this._activeWorkspaceY = 0; // y offset of active ws while dragging
         this._scroll = null;
@@ -663,13 +694,7 @@ SingleView.prototype = {
         this._animatingScroll = false; // programatically move the scroll bar
         this._inDrag = false; // dragging a window
         this._lastMotionTime = -1; // used to track "stopping" while dragging workspaces
-        let primary = global.get_primary_monitor();
-        this._dropGroup = new Clutter.Group({ x: 0, y: 0,
-                                              width: primary.width,
-                                              height: primary.height });
-        this._dropGroup._delegate = this;
-        global.stage.add_actor(this._dropGroup);
-        this._dropGroup.lower_bottom();
+
         this._dragIndex = -1;
 
         this._buttonPressId = 0;
@@ -769,6 +794,7 @@ SingleView.prototype = {
     _scrollToActive: function(showAnimation) {
         let active = global.screen.get_active_workspace_index();
 
+        this._computeWorkspacePositions();
         this._updateWorkspaceActors(showAnimation);
         this._scrollScrollBarToIndex(active, showAnimation);
     },
@@ -1085,11 +1111,11 @@ SingleView.prototype = {
 
         if (newNumWorkspaces > oldNumWorkspaces) {
             for (let w = oldNumWorkspaces; w < newNumWorkspaces; w++) {
-                this._actor.add_actor(this._workspaces[w].actor);
+                this.actor.add_actor(this._workspaces[w].actor);
                 this._workspaces[w]._windowDragBeginId = this._workspaces[w].connect('window-drag-begin',
-                                                                                     Lang.bind(this, this._onWindowDragBegin));
+                                                                                     Lang.bind(this, this._dragBegin));
                 this._workspaces[w]._windowDragEndId = this._workspaces[w].connect('window-drag-end',
-                                                                                   Lang.bind(this, this._onWindowDragEnd));
+                                                                                   Lang.bind(this, this._dragEnd));
             }
 
             this._computeWorkspacePositions();
@@ -1114,11 +1140,18 @@ SingleView.prototype = {
     _onDestroy: function() {
         GenericWorkspacesView.prototype._onDestroy.call(this);
         this._setWorkspaceDraggable(this._dragIndex, false);
-        this._dropGroup.destroy();
         if (this._timeoutId) {
             Mainloop.source_remove(this._timeoutId);
             this._timeoutId = 0;
         }
+        if (this._itemDragBeginId > 0) {
+            Main.overview.disconnect(this._itemDragBeginId);
+            this._itemDragBeginId = 0;
+        }
+        if (this._itemDragEndId > 0) {
+            Main.overview.disconnect(this._itemDragEndId);
+            this._itemDragEndId = 0;
+        }
         for (let w = 0; w < this._workspaces.length; w++) {
             if (this._workspaces[w]._windowDragBeginId) {
                 this._workspaces[w].disconnect(this._workspaces[w]._windowDragBeginId);
@@ -1131,44 +1164,7 @@ SingleView.prototype = {
         }
     },
 
-    acceptDrop: function(source, dropActor, x, y, time) {
-        if (x < this._x || y < this._y || y > this._y + this._height) {
-            this._dropGroup.lower_bottom();
-            dropActor.hide();
-            let target = global.stage.get_actor_at_pos(Clutter.PickMode.ALL, x, y);
-            dropActor.show();
-
-            if (target._delegate && target._delegate != this && target._delegate.acceptDrop) {
-                let [targX, targY] = target.get_transformed_position();
-                return target._delegate.acceptDrop(source, dropActor,
-                                                   (x - targX) / target.scale_x,
-                                                   (y - targY) / target.scale_y,
-                                                   time);
-            }
-            return false;
-        }
-
-        for (let i = 0; i < this._workspaces.length; i++) {
-            let [dx, dy] = this._workspaces[i].actor.get_transformed_position();
-            let [dw, dh] = this._workspaces[i].actor.get_transformed_size();
-
-            if (x > dx && x < dx + dw && y > dy && y < dy + dh) {
-                let accept = this._workspaces[i].acceptDrop(source, dropActor, x, y, time);
-                if (accept)
-                    this._workspaces[i].metaWorkspace.activate(time);
-                return accept;
-            }
-        }
-
-        let [dx, dy] = this._newWorkspaceArea.actor.get_transformed_position();
-        let [dw, dh] = this._newWorkspaceArea.actor.get_transformed_size();
-        if (x > dx && y > dy && y < dy + dh)
-            return this._acceptNewWorkspaceDrop(source, dropActor, x, y, time);
-
-        return false;
-    },
-
-    _onWindowDragBegin: function(w, actor) {
+    _dragBegin: function() {
         if (!this._scroll || this._scrolling)
             return;
 
@@ -1176,77 +1172,92 @@ SingleView.prototype = {
         this._computeWorkspacePositions();
         this._updateWorkspaceActors(true);
 
-        this._dropGroup.raise_top();
+        this._dragMonitor = {
+            dragMotion: Lang.bind(this, this._onDragMotion)
+        };
+        DND.addDragMonitor(this._dragMonitor);
     },
 
-    handleDragOver: function(self, actor, x, y) {
-        let onPanel = false;
+    _onDragMotion: function(dragEvent) {
+        let primary = global.get_primary_monitor();
 
         let activeWorkspaceIndex = global.screen.get_active_workspace_index();
-        if (x == 0 && activeWorkspaceIndex > 0 && this._dragOverLastX !== 0) {
-            this._workspaces[activeWorkspaceIndex - 1].metaWorkspace.activate(global.get_current_time());
-            this._workspaces[activeWorkspaceIndex - 1].setReservedSlot(actor._delegate);
-            this._dragOverLastX = 0;
-            return;
+        let leftWorkspace  = this._workspaces[activeWorkspaceIndex - 1];
+        let rightWorkspace = this._workspaces[activeWorkspaceIndex + 1];
+        let hoverWorkspace = null;
+
+        // reactive monitor edges
+        let leftEdge = primary.x;
+        let switchLeft = (dragEvent.x <= leftEdge && leftWorkspace);
+        if (switchLeft && this._dragOverLastX != leftEdge) {
+            leftWorkspace.metaWorkspace.activate(global.get_current_time());
+            leftWorkspace.setReservedSlot(dragEvent.dragActor._delegate);
+            this._dragOverLastX = leftEdge;
+
+            return DND.DragMotionResult.MOVE_DROP;
         }
-        if (x == global.screen_width - 1 && this._workspaces[activeWorkspaceIndex + 1] &&
-            this._dragOverLastX != global.screen_width - 1) {
-            this._workspaces[activeWorkspaceIndex + 1].metaWorkspace.activate(global.get_current_time());
-            this._workspaces[activeWorkspaceIndex + 1].setReservedSlot(actor._delegate);
-            this._dragOverLastX = global.screen_width - 1;
-            return;
+        let rightEdge = primary.x + primary.width - 1;
+        let switchRight = (dragEvent.x >= rightEdge && rightWorkspace);
+        if (switchRight && this._dragOverLastX != rightEdge) {
+            rightWorkspace.metaWorkspace.activate(global.get_current_time());
+            rightWorkspace.setReservedSlot(dragEvent.dragActor._delegate);
+            this._dragOverLastX = rightEdge;
+
+            return DND.DragMotionResult.MOVE_DROP;
+        }
+        this._dragOverLastX = dragEvent.x;
+
+        // check hover state of new workspace area / inactive workspaces
+        if (leftWorkspace) {
+            if (dragEvent.targetActor == this._leftShadow) {
+                hoverWorkspace = leftWorkspace;
+                leftWorkspace.opacity = leftWorkspace.actor.opacity = 255;
+            } else {
+                leftWorkspace.opacity = leftWorkspace.actor.opacity = 200;
+            }
         }
 
-        this._dragOverLastX = x;
-
-        let [dx, dy] = this._newWorkspaceArea.actor.get_transformed_position();
-        let [dw, dh] = this._newWorkspaceArea.actor.get_transformed_size();
-        this._newWorkspaceArea.setStyle(x > dx && y > dy && y < dy + dh);
-
-        [dx, dy] = this._leftShadow.get_transformed_position();
-        [dw, dh] = this._leftShadow.get_transformed_size();
-        if (this._workspaces[activeWorkspaceIndex - 1]) {
-            if (x > dx && x < dx + dw && y > dy && y < dy + dh) {
-                onPanel = -1;
-                this._workspaces[activeWorkspaceIndex - 1].actor.opacity = 255;
-            } else
-                this._workspaces[activeWorkspaceIndex - 1].actor.opacity = 200;
+        if (rightWorkspace) {
+            if (dragEvent.targetActor == this._rightShadow) {
+                hoverWorkspace = rightWorkspace;
+                rightWorkspace.opacity = rightWorkspace.actor.opacity = 255;
+            } else {
+                rightWorkspace.opacity = rightWorkspace.actor.opacity = 200;
+            }
+        } else {
+            let targetParent = dragEvent.targetActor.get_parent();
+            if (targetParent == this._newWorkspaceArea.actor)
+                this._newWorkspaceArea.setStyle(true);
+            else
+                this._newWorkspaceArea.setStyle(false);
         }
 
-        [dx, dy] = this._rightShadow.get_transformed_position();
-        [dw, dh] = this._rightShadow.get_transformed_size();
-        if (this._workspaces[activeWorkspaceIndex + 1]) {
-            if (x > dx && x < dx + dw && y > dy && y < dy + dh) {
-                onPanel = 1;
-                this._workspaces[activeWorkspaceIndex + 1].actor.opacity = 255;
-            } else
-                this._workspaces[activeWorkspaceIndex + 1].actor.opacity = 200;
-        }
-        if (onPanel) {
+
+        // handle delayed workspace switches
+        if (hoverWorkspace) {
             if (!this._timeoutId)
-                this._timeoutId = Mainloop.timeout_add_seconds (1, Lang.bind(this, function() {
-                   let i = global.screen.get_active_workspace_index();
-                   if (this._workspaces[i + onPanel]) {
-                       this._workspaces[i + onPanel].metaWorkspace.activate(global.get_current_time());
-                       this._workspaces[i + onPanel].setReservedSlot(actor._delegate);
-                   }
-                   return true;
-                }));
+                this._timeoutId = Mainloop.timeout_add_seconds(1,
+                    Lang.bind(this, function() {
+                       hoverWorkspace.metaWorkspace.activate(global.get_current_time());
+                       hoverWorkspace.setReservedSlot(dragEvent.dragActor._delegate);
+                       return false;
+                    }));
         } else {
             if (this._timeoutId) {
                 Mainloop.source_remove(this._timeoutId);
                 this._timeoutId = 0;
             }
         }
+
+        return DND.DragMotionResult.CONTINUE;
     },
 
-    _onWindowDragEnd: function(w, actor) {
+    _dragEnd: function() {
         if (this._timeoutId) {
             Mainloop.source_remove(this._timeoutId);
             this._timeoutId = 0;
         }
-        this._dropGroup.lower_bottom();
-        actor.opacity = 255;
+        DND.removeMonitor(this._dragMonitor);
         this._inDrag = false;
         this._computeWorkspacePositions();
         this._updateWorkspaceActors(true);
@@ -1418,9 +1429,7 @@ WorkspacesControls.prototype = {
         this.actor = new St.BoxLayout({ style_class: 'workspaces-bar' });
         this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
 
-        this._gconf = Shell.GConf.get_default();
-
-        let view = this._gconf.get_string(WORKSPACES_VIEW_KEY).toUpperCase();
+        let view = global.settings.get_string(WORKSPACES_VIEW_KEY).toUpperCase();
         if (view in WorkspacesViewType)
             this._currentViewType = WorkspacesViewType[view];
         else
@@ -1517,7 +1526,7 @@ WorkspacesControls.prototype = {
             this._toggleViewButton.set_style_class_name('workspace-controls switch-single');
 
         this._currentViewType = view;
-        this._gconf.set_string(WORKSPACES_VIEW_KEY, view);
+        global.settings.set_string(WORKSPACES_VIEW_KEY, view);
     },
 
     _onDestroy: function() {
@@ -1579,8 +1588,8 @@ WorkspacesManager.prototype = {
         this.controlsBar.actor.connect('destroy',
                                        Lang.bind(this, this._onDestroy));
         this._viewChangedId =
-            Shell.GConf.get_default().connect('changed::' + WORKSPACES_VIEW_KEY,
-                                              Lang.bind(this, this._updateView));
+            global.settings.connect('changed::' + WORKSPACES_VIEW_KEY,
+                                    Lang.bind(this, this._updateView));
         this._nWorkspacesNotifyId =
             global.screen.connect('notify::n-workspaces',
                                   Lang.bind(this, this._workspacesChanged));
@@ -1589,7 +1598,7 @@ WorkspacesManager.prototype = {
     _updateView: function() {
         let viewType, newView;
 
-        let view = Shell.GConf.get_default().get_string(WORKSPACES_VIEW_KEY).toUpperCase();
+        let view = global.settings.get_string(WORKSPACES_VIEW_KEY).toUpperCase();
         if (view in WorkspacesViewType)
             viewType = WorkspacesViewType[view];
         else
@@ -1667,7 +1676,7 @@ WorkspacesManager.prototype = {
         if (this._nWorkspacesNotifyId > 0)
             global.screen.disconnect(this._nWorkspacesNotifyId);
         if (this._viewChangedId > 0)
-            Shell.GConf.get_default().disconnect(this._viewChangedId);
+            global.settings.disconnect(this._viewChangedId);
         for (let w = 0; w < this._workspaces.length; w++) {
             this._workspaces[w].disconnectAll();
             this._workspaces[w].destroy();
