@@ -8,8 +8,6 @@
 #include <gio/gio.h>
 #include <gio/gdesktopappinfo.h>
 #include <gtk/gtk.h>
-#include <gconf/gconf.h>
-#include <gconf/gconf-client.h>
 #include <clutter/clutter.h>
 
 #include "shell-app-private.h"
@@ -20,14 +18,13 @@
 #define GMENU_I_KNOW_THIS_IS_UNSTABLE
 #include <gmenu-tree.h>
 
-#define SHELL_APP_FAVORITES_KEY "/desktop/gnome/shell/favorite_apps"
-
 /* Vendor prefixes are something that can be preprended to a .desktop
  * file name.  Undo this.
  */
 static const char*const known_vendor_prefixes[] = { "gnome",
                                                     "fedora",
-                                                    "mozilla" };
+                                                    "mozilla",
+                                                    NULL };
 
 enum {
    PROP_0,
@@ -196,7 +193,6 @@ static void
 shell_app_system_init (ShellAppSystem *self)
 {
   ShellAppSystemPrivate *priv;
-  GConfClient *client;
 
   self->priv = priv = G_TYPE_INSTANCE_GET_PRIVATE (self,
                                                    SHELL_TYPE_APP_SYSTEM,
@@ -222,8 +218,6 @@ shell_app_system_init (ShellAppSystem *self)
   gmenu_tree_add_monitor (priv->settings_tree, on_tree_changed_cb, self);
 
   reread_menus (self);
-
-  client = gconf_client_get_default ();
 }
 
 static void
@@ -308,9 +302,15 @@ reread_entries (ShellAppSystem     *self,
   g_slist_free (*cache);
   *cache = NULL;
 
-  *cache = gather_entries_recurse (self, *cache, unique, trunk);
-
-  gmenu_tree_item_unref (trunk);
+  if (!trunk)
+    {
+      *cache = NULL;
+    }
+  else
+    {
+      *cache = gather_entries_recurse (self, *cache, unique, trunk);
+      gmenu_tree_item_unref (trunk);
+    }
 }
 
 static void
@@ -484,6 +484,49 @@ shell_app_system_get_app (ShellAppSystem   *self,
   return app;
 }
 
+/**
+ * shell_app_system_get_app_for_path:
+ * @system: a #ShellAppSystem
+ * @desktop_path: (utf8): UTF-8 encoded absolute file name
+ *
+ * Find or create a #ShellApp corresponding to a given absolute
+ * file name which must be in the standard paths (XDG_DATA_DIRS).
+ * For files outside the datadirs, this function returns %NULL.
+ *
+ * If already cached elsewhere in memory, return that instance.
+ * Otherwise, create a new one.
+ *
+ * Return value: (transfer full): The #ShellApp for id, or %NULL if none
+ */
+ShellApp *
+shell_app_system_get_app_for_path (ShellAppSystem   *system,
+                                   const char       *desktop_path)
+{
+  const char *basename;
+  ShellAppInfo *info;
+
+  basename = g_strrstr (desktop_path, "/");
+  if (basename)
+    basename += 1;
+  else
+    basename = desktop_path;
+
+  info = g_hash_table_lookup (system->priv->app_id_to_info, basename);
+  if (!info)
+    return NULL;
+
+  if (info->type == SHELL_APP_INFO_TYPE_ENTRY)
+    {
+      const char *full_path = gmenu_tree_entry_get_desktop_file_path ((GMenuTreeEntry*) info->entry);
+      if (strcmp (desktop_path, full_path) != 0)
+        return NULL;
+    }
+  else
+    return NULL;
+
+  return shell_app_system_get_app (system, basename);
+}
+
 /**
  * shell_app_system_get_app_for_window:
  * @self: A #ShellAppSystem
@@ -1059,10 +1102,84 @@ shell_app_info_get_icon (ShellAppInfo *info)
   return NULL;
 }
 
-GSList *
-shell_app_info_get_categories (ShellAppInfo *info)
+/**
+ * shell_app_system_get_sections:
+ *
+ * return names of sections in applications menu.
+ *
+ * Returns: (element-type utf8) (transfer full): List of Names
+ */
+GList *
+shell_app_system_get_sections (ShellAppSystem *system)
 {
-  return NULL; /* TODO */
+  GList *res = NULL;
+  GSList *i, *contents;
+  GMenuTreeDirectory *root;
+
+  root = gmenu_tree_get_root_directory (system->priv->apps_tree);
+
+  if (G_UNLIKELY (!root))
+    g_error ("applications.menu not found.");
+
+  contents = gmenu_tree_directory_get_contents (root);
+
+  for (i = contents; i; i = i->next)
+    {
+      GMenuTreeItem *item = i->data;
+      if (gmenu_tree_item_get_type (item) == GMENU_TREE_ITEM_DIRECTORY)
+        {
+          char *name = g_strdup (gmenu_tree_directory_get_name ((GMenuTreeDirectory*)item));
+
+          g_assert (name);
+
+          res = g_list_append (res, name);
+        }
+      gmenu_tree_item_unref (item);
+    }
+
+  g_slist_free (contents);
+
+  return res;
+}
+
+/**
+ * shell_app_info_get_section:
+ *
+ * return name of section, that contain this application.
+ * Returns: (transfer full): section name
+ */
+char *
+shell_app_info_get_section (ShellAppInfo *info)
+{
+  char *name;
+  GMenuTreeDirectory *dir, *parent;
+
+  if (info->type != SHELL_APP_INFO_TYPE_ENTRY)
+    return NULL;
+
+  dir = gmenu_tree_item_get_parent ((GMenuTreeItem*)info->entry);
+  if (!dir)
+    return NULL;
+
+  parent = gmenu_tree_item_get_parent ((GMenuTreeItem*)dir);
+  if (!parent)
+    return NULL;
+
+  while (TRUE)
+    {
+      GMenuTreeDirectory *pparent = gmenu_tree_item_get_parent ((GMenuTreeItem*)parent);
+      if (!pparent)
+        break;
+      gmenu_tree_item_unref ((GMenuTreeItem*)dir);
+      dir = parent;
+      parent = pparent;
+    }
+
+  name = g_strdup (gmenu_tree_directory_get_name (dir));
+
+  gmenu_tree_item_unref ((GMenuTreeItem*)dir);
+  gmenu_tree_item_unref ((GMenuTreeItem*)parent);
+  return name;
 }
 
 gboolean
@@ -1181,10 +1298,16 @@ shell_app_info_launch_full (ShellAppInfo *info,
       meta_window_activate (info->window, timestamp);
       return TRUE;
     }
-
-  filename = shell_app_info_get_desktop_file_path (info);
-  gapp = g_desktop_app_info_new_from_filename (filename);
-  g_free (filename);
+  else if (info->type == SHELL_APP_INFO_TYPE_ENTRY)
+    {
+      gapp = g_desktop_app_info_new (shell_app_info_get_id (info));
+    }
+  else
+    {
+      filename = shell_app_info_get_desktop_file_path (info);
+      gapp = g_desktop_app_info_new_from_filename (filename);
+      g_free (filename);
+    }
 
   if (!gapp)
     {
@@ -1199,11 +1322,6 @@ shell_app_info_launch_full (ShellAppInfo *info,
   if (timestamp == 0)
     timestamp = clutter_get_current_event_time ();
 
-  /* Shell design calls for on application launch, no window is focused,
-   * and we have startup notification displayed.
-   */
-  meta_display_focus_the_no_focus_window (display, screen, timestamp);
-
   if (workspace < 0)
     workspace = meta_screen_get_active_workspace_index (screen);
 
