@@ -1,9 +1,13 @@
 /* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
 
-#include "shell-global.h"
+#include "config.h"
+
+#include "shell-global-private.h"
+#include "shell-perf-log.h"
 #include "shell-wm.h"
 
 #include "display.h"
+#include "util.h"
 #include <clutter/glx/clutter-glx.h>
 #include <clutter/x11/clutter-x11.h>
 #include <gdk/gdkx.h>
@@ -15,13 +19,18 @@
 #include <unistd.h>
 #include <dbus/dbus-glib.h>
 #include <gio/gio.h>
-#include <glib/gi18n-lib.h>
 #include <math.h>
 #include <X11/extensions/Xfixes.h>
+#include <gjs/gjs.h>
+#ifdef HAVE_SYS_RESOURCE_H
+#include <sys/resource.h>
+#endif
 
 #define SHELL_DBUS_SERVICE "org.gnome.Shell"
+#define MAGNIFIER_DBUS_SERVICE "org.gnome.Magnifier"
 
 static void grab_notify (GtkWidget *widget, gboolean is_grab, gpointer user_data);
+static void update_root_window_pixmap (ShellGlobal *global);
 
 struct _ShellGlobal {
   GObject parent;
@@ -37,15 +46,23 @@ struct _ShellGlobal {
 
   ShellStageInputMode input_mode;
   XserverRegion input_region;
-  
+
+  GjsContext *js_context;
   MutterPlugin *plugin;
   ShellWM *wm;
+  GSettings *settings;
   const char *datadir;
   const char *imagedir;
-  const char *configdir;
+  const char *userdatadir;
 
   /* Displays the root window; see shell_global_create_root_pixmap_actor() */
   ClutterActor *root_pixmap;
+
+  gint last_change_screen_width, last_change_screen_height;
+
+  guint work_count;
+  GSList *leisure_closures;
+  guint leisure_function_id;
 };
 
 enum {
@@ -53,28 +70,20 @@ enum {
 
   PROP_OVERLAY_GROUP,
   PROP_SCREEN,
+  PROP_GDK_SCREEN,
   PROP_SCREEN_WIDTH,
   PROP_SCREEN_HEIGHT,
   PROP_STAGE,
   PROP_WINDOW_GROUP,
   PROP_WINDOW_MANAGER,
+  PROP_SETTINGS,
   PROP_DATADIR,
   PROP_IMAGEDIR,
-  PROP_CONFIGDIR,
-};
-
-/* Signals */
-enum
-{
-  PANEL_RUN_DIALOG,
-  PANEL_MAIN_MENU,
-  LAST_SIGNAL
+  PROP_USERDATADIR,
 };
 
 G_DEFINE_TYPE(ShellGlobal, shell_global, G_TYPE_OBJECT);
 
-static guint shell_global_signals [LAST_SIGNAL] = { 0 };
-
 static void
 shell_global_set_property(GObject         *object,
                           guint            prop_id,
@@ -105,6 +114,9 @@ shell_global_get_property(GObject         *object,
     case PROP_SCREEN:
       g_value_set_object (value, shell_global_get_screen (global));
       break;
+    case PROP_GDK_SCREEN:
+      g_value_set_object (value, shell_global_get_gdk_screen (global));
+      break;
     case PROP_SCREEN_WIDTH:
       {
         int width, height;
@@ -130,14 +142,17 @@ shell_global_get_property(GObject         *object,
     case PROP_WINDOW_MANAGER:
       g_value_set_object (value, global->wm);
       break;
+    case PROP_SETTINGS:
+      g_value_set_object (value, global->settings);
+      break;
     case PROP_DATADIR:
       g_value_set_string (value, global->datadir);
       break;
     case PROP_IMAGEDIR:
       g_value_set_string (value, global->imagedir);
       break;
-    case PROP_CONFIGDIR:
-      g_value_set_string (value, global->configdir);
+    case PROP_USERDATADIR:
+      g_value_set_string (value, global->userdatadir);
       break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -150,7 +165,6 @@ shell_global_init (ShellGlobal *global)
 {
   const char *datadir = g_getenv ("GNOME_SHELL_DATADIR");
   char *imagedir;
-  GFile *conf_dir;
 
   if (!datadir)
     datadir = GNOME_SHELL_DATADIR;
@@ -170,10 +184,10 @@ shell_global_init (ShellGlobal *global)
     }
 
   /* Ensure config dir exists for later use */
-  global->configdir = g_build_filename (g_get_home_dir (), ".gnome2", "shell", NULL);
-  conf_dir = g_file_new_for_path (global->configdir);
-  g_file_make_directory (conf_dir, NULL, NULL);
-  g_object_unref (conf_dir);
+  global->userdatadir = g_build_filename (g_get_user_data_dir (), "gnome-shell", NULL);
+  g_mkdir_with_parents (global->userdatadir, 0700);
+
+  global->settings = g_settings_new ("org.gnome.shell");
   
   global->grab_notifier = GTK_WINDOW (gtk_window_new (GTK_WINDOW_TOPLEVEL));
   g_signal_connect (global->grab_notifier, "grab-notify", G_CALLBACK (grab_notify), global);
@@ -182,6 +196,9 @@ shell_global_init (ShellGlobal *global)
   global->root_pixmap = NULL;
 
   global->input_mode = SHELL_STAGE_INPUT_MODE_NORMAL;
+
+  global->last_change_screen_width = 0;
+  global->last_change_screen_height = 0;
 }
 
 static void
@@ -192,24 +209,6 @@ shell_global_class_init (ShellGlobalClass *klass)
   gobject_class->get_property = shell_global_get_property;
   gobject_class->set_property = shell_global_set_property;
 
-  shell_global_signals[PANEL_RUN_DIALOG] =
-    g_signal_new ("panel-run-dialog",
-		  G_TYPE_FROM_CLASS (klass),
-		  G_SIGNAL_RUN_LAST,
-		  G_STRUCT_OFFSET (ShellGlobalClass, panel_run_dialog),
-		  NULL, NULL,
-		  g_cclosure_marshal_VOID__INT,
-		  G_TYPE_NONE, 1, G_TYPE_INT);
-
-  shell_global_signals[PANEL_MAIN_MENU] =
-    g_signal_new ("panel-main-menu",
-		  G_TYPE_FROM_CLASS (klass),
-		  G_SIGNAL_RUN_LAST,
-		  G_STRUCT_OFFSET (ShellGlobalClass, panel_main_menu),
-		  NULL, NULL,
-		  g_cclosure_marshal_VOID__INT,
-		  G_TYPE_NONE, 1, G_TYPE_INT);
-
   g_object_class_install_property (gobject_class,
                                    PROP_OVERLAY_GROUP,
                                    g_param_spec_object ("overlay-group",
@@ -224,6 +223,15 @@ shell_global_class_init (ShellGlobalClass *klass)
                                                         "Metacity screen object for the shell",
                                                         META_TYPE_SCREEN,
                                                         G_PARAM_READABLE));
+
+  g_object_class_install_property (gobject_class,
+                                   PROP_GDK_SCREEN,
+                                   g_param_spec_object ("gdk-screen",
+                                                        "GdkScreen",
+                                                        "Gdk screen object for the shell",
+                                                        GDK_TYPE_SCREEN,
+                                                        G_PARAM_READABLE));
+
   g_object_class_install_property (gobject_class,
                                    PROP_SCREEN_WIDTH,
                                    g_param_spec_int ("screen-width",
@@ -260,6 +268,13 @@ shell_global_class_init (ShellGlobalClass *klass)
                                                         "Window management interface",
                                                         SHELL_TYPE_WM,
                                                         G_PARAM_READABLE));
+  g_object_class_install_property (gobject_class,
+                                   PROP_SETTINGS,
+                                   g_param_spec_object ("settings",
+                                                        "Settings",
+                                                        "GSettings instance for gnome-shell configuration",
+                                                        G_TYPE_SETTINGS,
+                                                        G_PARAM_READABLE));
   g_object_class_install_property (gobject_class,
                                    PROP_DATADIR,
                                    g_param_spec_string ("datadir",
@@ -275,10 +290,10 @@ shell_global_class_init (ShellGlobalClass *klass)
                                                         NULL,
                                                         G_PARAM_READABLE));
   g_object_class_install_property (gobject_class,
-                                   PROP_CONFIGDIR,
-                                   g_param_spec_string ("configdir",
-                                                        "Configuration directory",
-                                                        "Directory containing gnome-shell configuration files",
+                                   PROP_USERDATADIR,
+                                   g_param_spec_string ("userdatadir",
+                                                        "User data directory",
+                                                        "Directory containing gnome-shell user data",
                                                         NULL,
                                                         G_PARAM_READABLE));
 }
@@ -419,6 +434,19 @@ shell_global_get_screen (ShellGlobal  *global)
   return mutter_plugin_get_screen (global->plugin);
 }
 
+/**
+ * shell_global_get_gdk_screen:
+ *
+ * Return value: (transfer none): Gdk screen object for the shell
+ */
+GdkScreen *
+shell_global_get_gdk_screen (ShellGlobal *global)
+{
+  g_return_val_if_fail (SHELL_IS_GLOBAL (global), NULL);
+
+  return gdk_screen_get_default ();
+}
+
 /**
  * shell_global_get_windows:
  *
@@ -434,15 +462,114 @@ shell_global_get_windows (ShellGlobal *global)
   return mutter_plugin_get_windows (global->plugin);
 }
 
+static gboolean
+update_screen_size (gpointer data)
+{
+  int width, height;
+  ShellGlobal *global = SHELL_GLOBAL (data);
+
+  mutter_plugin_query_screen_size (global->plugin, &width, &height);
+
+  if (global->last_change_screen_width == width && global->last_change_screen_height == height)
+    return FALSE;
+
+  global->last_change_screen_width = width;
+  global->last_change_screen_height = height;
+
+  /* update size of background actor to fix tiled backgrounds */
+  if (global->root_pixmap)
+    clutter_actor_set_size (CLUTTER_ACTOR (global->root_pixmap),
+                            width, height);
+
+  return FALSE;
+}
+
+static void
+global_stage_notify_width (GObject    *gobject,
+                           GParamSpec *pspec,
+                           gpointer    data)
+{
+  ShellGlobal *global = SHELL_GLOBAL (data);
+
+  g_object_notify (G_OBJECT (global), "screen-width");
+
+  meta_later_add (META_LATER_BEFORE_REDRAW,
+                  update_screen_size,
+                  global,
+                  NULL);
+}
+
+static void
+global_stage_notify_height (GObject    *gobject,
+                            GParamSpec *pspec,
+                            gpointer    data)
+{
+  ShellGlobal *global = SHELL_GLOBAL (data);
+
+  g_object_notify (G_OBJECT (global), "screen-height");
+
+  meta_later_add (META_LATER_BEFORE_REDRAW,
+                  update_screen_size,
+                  global,
+                  NULL);
+}
+
+static void
+global_stage_before_paint (ClutterStage *stage,
+                           ShellGlobal  *global)
+{
+  shell_perf_log_event (shell_perf_log_get_default (),
+                        "clutter.stagePaintStart");
+}
+
+static void
+global_stage_after_paint (ClutterStage *stage,
+                          ShellGlobal  *global)
+{
+  shell_perf_log_event (shell_perf_log_get_default (),
+                        "clutter.stagePaintDone");
+}
+
 void
 _shell_global_set_plugin (ShellGlobal  *global,
                           MutterPlugin *plugin)
 {
+  ClutterActor *stage;
+
   g_return_if_fail (SHELL_IS_GLOBAL (global));
   g_return_if_fail (global->plugin == NULL);
 
   global->plugin = plugin;
   global->wm = shell_wm_new (plugin);
+
+  stage = mutter_plugin_get_stage (plugin);
+
+  g_signal_connect (stage, "notify::width",
+                    G_CALLBACK (global_stage_notify_width), global);
+  g_signal_connect (stage, "notify::height",
+                    G_CALLBACK (global_stage_notify_height), global);
+  update_screen_size (global);
+
+  g_signal_connect (stage, "paint",
+                    G_CALLBACK (global_stage_before_paint), global);
+  g_signal_connect_after (stage, "paint",
+                          G_CALLBACK (global_stage_after_paint), global);
+
+  shell_perf_log_define_event (shell_perf_log_get_default(),
+                               "clutter.stagePaintStart",
+                               "Start of stage page repaint",
+                               "");
+  shell_perf_log_define_event (shell_perf_log_get_default(),
+                               "clutter.stagePaintDone",
+                               "End of stage page repaint",
+                               "");
+}
+
+void
+_shell_global_set_gjs_context (ShellGlobal *global,
+                               GjsContext  *context)
+{
+  global->js_context = context;
 }
 
 /**
@@ -508,6 +635,64 @@ shell_global_display_is_grabbed (ShellGlobal *global)
   return meta_display_get_grab_op (display) != META_GRAB_OP_NONE;
 }
 
+/**
+ * shell_global_add_extension_importer:
+ * @target_object_script: JavaScript code evaluating to a target object
+ * @target_property: Name of property to use for importer
+ * @directory: Source directory:
+ * @error: A #GError
+ *
+ * This function sets a property named @target_property on the object
+ * resulting from the evaluation of @target_object_script code, which
+ * acts as a GJS importer for directory @directory.
+ *
+ * Returns: %TRUE on success
+ */
+gboolean
+shell_global_add_extension_importer (ShellGlobal *global,
+                                     const char  *target_object_script,
+                                     const char  *target_property,
+                                     const char  *directory,
+                                     GError     **error)
+{
+  jsval target_object;
+  JSObject *importer;
+  JSContext *context = gjs_context_get_native_context (global->js_context);
+  char *search_path[2] = { 0, 0 };
+
+  // This is a bit of a hack; ideally we'd be able to pass our target
+  // object directly into this function, but introspection doesn't
+  // support that at the moment.  Instead evaluate a string to get it.
+  if (!JS_EvaluateScript(context,
+                         JS_GetGlobalObject(context),
+                         target_object_script,
+                         strlen (target_object_script),
+                         "<target_object_script>",
+                         0,
+                         &target_object))
+    {
+      char *message;
+      gjs_log_exception(context,
+                        &message);
+      g_set_error(error,
+                  G_IO_ERROR,
+                  G_IO_ERROR_FAILED,
+                  "%s", message ? message : "(unknown)");
+      g_free(message);
+      return FALSE;
+    }
+
+  if (!JSVAL_IS_OBJECT (target_object))
+    {
+      g_error ("shell_global_add_extension_importer: invalid target object");
+      return FALSE;
+    }
+
+  search_path[0] = (char*)directory;
+  importer = gjs_define_importer (context, JSVAL_TO_OBJECT (target_object), target_property, (const char **)search_path, FALSE);
+  return TRUE;
+}
+
 /* Code to close all file descriptors before we exec; copied from gspawn.c in GLib.
  *
  * Authors: Padraig O'Briain, Matthias Clasen, Lennart Poettering
@@ -640,7 +825,51 @@ shell_global_reexec_self (ShellGlobal *global)
   g_ptr_array_free (arr, TRUE);
 }
 
-void 
+/**
+ * shell_global_breakpoint:
+ * @global: A #ShellGlobal
+ *
+ * Using G_BREAKPOINT(), interrupt the current process.  This is useful
+ * in conjunction with a debugger such as gdb.
+ */
+void
+shell_global_breakpoint (ShellGlobal *global)
+{
+  G_BREAKPOINT ();
+}
+
+/**
+ * shell_global_gc:
+ * @global: A #ShellGlobal
+ *
+ * Start a garbage collection process.  For more information, see
+ * https://developer.mozilla.org/En/JS_GC
+ */
+void
+shell_global_gc (ShellGlobal *global)
+{
+  JSContext *context = gjs_context_get_native_context (global->js_context);
+
+  JS_GC (context);
+}
+
+/**
+ * shell_global_maybe_gc:
+ * @global: A #ShellGlobal
+ *
+ * Start a garbage collection process when it would free up enough memory
+ * to be worth the amount of time it would take
+ * https://developer.mozilla.org/en/SpiderMonkey/JSAPI_Reference/JS_MaybeGC
+ */
+void
+shell_global_maybe_gc (ShellGlobal *global)
+{
+  JSContext *context = gjs_context_get_native_context (global->js_context);
+
+  JS_MaybeGC (context);
+}
+
+void
 shell_global_grab_dbus_service (ShellGlobal *global)
 {
   GError *error = NULL;
@@ -673,7 +902,7 @@ shell_global_grab_dbus_service (ShellGlobal *global)
    * unless a special environment variable is passed.  The environment
    * variable is used by the gnome-shell (no --replace) launcher in
    * Xephyr */
-  if (!g_getenv ("GNOME_SHELL_NO_REPLACE_PANEL"))
+  if (!g_getenv ("GNOME_SHELL_NO_REPLACE"))
     {
       if (!dbus_g_proxy_call (bus, "RequestName", &error, G_TYPE_STRING,
                               "org.gnome.Panel", G_TYPE_UINT,
@@ -685,10 +914,24 @@ shell_global_grab_dbus_service (ShellGlobal *global)
           exit (1);
         }
     }
+
+  /* ...and the org.gnome.Magnifier service.
+   */
+  if (!dbus_g_proxy_call (bus, "RequestName", &error,
+                          G_TYPE_STRING, MAGNIFIER_DBUS_SERVICE,
+                          G_TYPE_UINT, 0,
+                          G_TYPE_INVALID,
+                          G_TYPE_UINT, &request_name_result,
+                          G_TYPE_INVALID))
+    {
+      g_print ("failed to acquire %s: %s\n", MAGNIFIER_DBUS_SERVICE, error->message);
+      /* Failing to acquire the magnifer service is not fatal.  Log the error,
+       * but keep going. */
+    }
   g_object_unref (bus);
 }
 
-static void 
+static void
 grab_notify (GtkWidget *widget, gboolean was_grabbed, gpointer user_data)
 {
   ShellGlobal *global = SHELL_GLOBAL (user_data);
@@ -757,22 +1000,6 @@ root_window_filter (GdkXEvent *native, GdkEvent *event, gpointer data)
   return GDK_FILTER_CONTINUE;
 }
 
-/* Workaround for a clutter bug where if ClutterGLXTexturePixmap
- * is painted without the pixmap being set, a crash will occur inside
- * Cogl.
- *
- * http://bugzilla.openedhand.com/show_bug.cgi?id=1644
- */
-static void
-root_pixmap_paint (ClutterActor *actor, gpointer data)
-{
-  Pixmap pixmap;
-
-  g_object_get (G_OBJECT (actor), "pixmap", &pixmap, NULL);
-  if (!pixmap)
-    g_signal_stop_emission_by_name (actor, "paint");
-}
-
 /*
  * Called when the root window pixmap actor is destroyed.
  */
@@ -812,16 +1039,24 @@ shell_global_format_time_relative_pretty (ShellGlobal *global,
     *text = g_strdup (_("Less than a minute ago"));
     *next_update = MINUTE - delta;
    } else if (delta < HOUR) {
-     *text = g_strdup_printf (ngettext ("%d minute ago", "%d minutes ago", delta / MINUTE), delta / MINUTE);
+     *text = g_strdup_printf (dngettext (GETTEXT_PACKAGE,
+                                         "%d minute ago", "%d minutes ago",
+                                         delta / MINUTE), delta / MINUTE);
      *next_update = MINUTE - (delta % MINUTE);
    } else if (delta < DAY) {
-     *text = g_strdup_printf (ngettext ("%d hour ago", "%d hours ago", delta / HOUR), delta / HOUR);
+     *text = g_strdup_printf (dngettext (GETTEXT_PACKAGE,
+                                         "%d hour ago", "%d hours ago",
+                                         delta / HOUR), delta / HOUR);
      *next_update = HOUR - (delta % HOUR);
    } else if (delta < WEEK) {
-     *text = g_strdup_printf (ngettext ("%d day ago", "%d days ago", delta / DAY), delta / DAY);
+     *text = g_strdup_printf (dngettext (GETTEXT_PACKAGE,
+                                         "%d day ago", "%d days ago",
+                                         delta / DAY), delta / DAY);
      *next_update = DAY - (delta % DAY);
    } else {
-     *text = g_strdup_printf (ngettext ("%d week ago", "%d weeks ago", delta / WEEK), delta / WEEK);
+     *text = g_strdup_printf (dngettext (GETTEXT_PACKAGE,
+                                         "%d week ago", "%d weeks ago",
+                                         delta / WEEK), delta / WEEK);
      *next_update = WEEK - (delta % WEEK);
    }
 }
@@ -840,6 +1075,7 @@ shell_global_create_root_pixmap_actor (ShellGlobal *global)
 {
   GdkWindow *window;
   ClutterActor *stage;
+  ClutterColor stage_color;
 
   /* The actor created is actually a ClutterClone of global->root_pixmap. */
 
@@ -847,20 +1083,36 @@ shell_global_create_root_pixmap_actor (ShellGlobal *global)
     {
       global->root_pixmap = clutter_glx_texture_pixmap_new ();
 
+      clutter_actor_set_size (CLUTTER_ACTOR (global->root_pixmap),
+                              global->last_change_screen_width,
+                              global->last_change_screen_height);
+
+      clutter_texture_set_repeat (CLUTTER_TEXTURE (global->root_pixmap),
+                                  TRUE, TRUE);
+
       /* The low and medium quality filters give nearest-neighbor resizing. */
       clutter_texture_set_filter_quality (CLUTTER_TEXTURE (global->root_pixmap),
                                           CLUTTER_TEXTURE_QUALITY_HIGH);
 
+      /* Initialize to the stage color, since that's what will be seen
+       * in the main view if there's no actual background window.
+       */
+      stage = mutter_plugin_get_stage (global->plugin);
+      clutter_stage_get_color (CLUTTER_STAGE (stage), &stage_color);
+      clutter_texture_set_from_rgb_data (CLUTTER_TEXTURE (global->root_pixmap),
+                                         /* ClutterColor has the same layout
+                                          * as one pixel of RGB(A) data.
+                                          */
+                                         (const guchar *)&stage_color, FALSE,
+                                         /* w, h, rowstride, bpp, flags */
+                                         1, 1, 3, 3, 0, NULL);
+
       /* We can only clone an actor within a stage, so we hide the source
        * texture then add it to the stage */
       clutter_actor_hide (global->root_pixmap);
-      stage = mutter_plugin_get_stage (global->plugin);
       clutter_container_add_actor (CLUTTER_CONTAINER (stage),
                                    global->root_pixmap);
 
-      g_signal_connect (global->root_pixmap, "paint",
-                        G_CALLBACK (root_pixmap_paint), NULL);
-
       /* This really should never happen; but just in case... */
       g_signal_connect (global->root_pixmap, "destroy",
                         G_CALLBACK (root_pixmap_destroy), global);
@@ -889,7 +1141,7 @@ shell_global_create_root_pixmap_actor (ShellGlobal *global)
  *
  * Gets a list of the bounding boxes of the active screen's monitors.
  *
- * Return value: (transfer full) (element-type GdkRectangle): a list
+ * Return value: (transfer full) (element-type Meta.Rectangle): a list
  * of monitor bounding boxes.
  */
 GSList *
@@ -900,17 +1152,11 @@ shell_global_get_monitors (ShellGlobal *global)
   MetaRectangle rect;
   int i;
 
-  g_assert (sizeof (MetaRectangle) == sizeof (GdkRectangle) &&
-            G_STRUCT_OFFSET (MetaRectangle, x) == G_STRUCT_OFFSET (GdkRectangle, x) &&
-            G_STRUCT_OFFSET (MetaRectangle, y) == G_STRUCT_OFFSET (GdkRectangle, y) &&
-            G_STRUCT_OFFSET (MetaRectangle, width) == G_STRUCT_OFFSET (GdkRectangle, width) &&
-            G_STRUCT_OFFSET (MetaRectangle, height) == G_STRUCT_OFFSET (GdkRectangle, height));
-
   for (i = meta_screen_get_n_monitors (screen) - 1; i >= 0; i--)
     {
       meta_screen_get_monitor_geometry (screen, i, &rect);
       monitors = g_slist_prepend (monitors,
-                                  g_boxed_copy (GDK_TYPE_RECTANGLE, &rect));
+                                  meta_rectangle_copy (&rect));
     }
   return monitors;
 }
@@ -924,20 +1170,49 @@ shell_global_get_monitors (ShellGlobal *global)
  *
  * Return value: the bounding box of the primary monitor
  */
-GdkRectangle *
+MetaRectangle *
 shell_global_get_primary_monitor (ShellGlobal  *global)
 {
-  MetaScreen *screen = shell_global_get_screen (global);
+  GdkScreen *screen = shell_global_get_gdk_screen (global);
+  GdkRectangle gdk_rect;
   MetaRectangle rect;
+  gint primary = 0;
 
-  g_assert (sizeof (MetaRectangle) == sizeof (GdkRectangle) &&
-            G_STRUCT_OFFSET (MetaRectangle, x) == G_STRUCT_OFFSET (GdkRectangle, x) &&
-            G_STRUCT_OFFSET (MetaRectangle, y) == G_STRUCT_OFFSET (GdkRectangle, y) &&
-            G_STRUCT_OFFSET (MetaRectangle, width) == G_STRUCT_OFFSET (GdkRectangle, width) &&
-            G_STRUCT_OFFSET (MetaRectangle, height) == G_STRUCT_OFFSET (GdkRectangle, height));
+  /* gdk_screen_get_primary_monitor is only present in gtk-2.20+
+   * and is in a useable state (supports heuristics and fallback modes)
+   * starting with 2.20.1
+   */
+#if !GTK_CHECK_VERSION (2, 20, 1)
+  gint i;
+  gchar *output_name = NULL;
+  gint num_monitors = gdk_screen_get_n_monitors (screen);
 
-  meta_screen_get_monitor_geometry (screen, 0, &rect);
-  return g_boxed_copy (GDK_TYPE_RECTANGLE, &rect);
+  for (i = 0; i < num_monitors; i++)
+    {
+      /* Prefer the laptop's internal screen if present */
+      output_name = gdk_screen_get_monitor_plug_name (screen, i);
+      if (output_name)
+        {
+          gboolean is_lvds = g_ascii_strncasecmp (output_name, "LVDS", 4) == 0;
+          g_free (output_name);
+          if (is_lvds)
+            {
+              primary = i;
+              break;
+            }
+        }
+    }
+#else
+  primary = gdk_screen_get_primary_monitor (screen);
+#endif
+
+  gdk_screen_get_monitor_geometry (screen, primary, &gdk_rect);
+  rect.x = gdk_rect.x;
+  rect.y = gdk_rect.y;
+  rect.width = gdk_rect.width;
+  rect.height = gdk_rect.height;
+
+  return meta_rectangle_copy (&rect);
 }
 
 /**
@@ -949,7 +1224,7 @@ shell_global_get_primary_monitor (ShellGlobal  *global)
  *
  * Return value: the bounding box of the focus monitor
  */
-GdkRectangle *
+MetaRectangle *
 shell_global_get_focus_monitor (ShellGlobal  *global)
 {
   MetaScreen *screen = shell_global_get_screen (global);
@@ -968,35 +1243,39 @@ shell_global_get_focus_monitor (ShellGlobal  *global)
         {
           meta_screen_get_monitor_geometry (screen, i, &rect);
 
-          if (rect.x < wrect.x && rect.y < wrect.y &&
+          if (rect.x <= wrect.x && rect.y <= wrect.y &&
               rect.x + rect.width > wrect.x &&
               rect.y + rect.height > wrect.y)
-            return g_boxed_copy (GDK_TYPE_RECTANGLE, &rect);
+            return meta_rectangle_copy (&rect);
         }
     }
 
   meta_screen_get_monitor_geometry (screen, 0, &rect);
-  return g_boxed_copy (GDK_TYPE_RECTANGLE, &rect);
+  return meta_rectangle_copy (&rect);
 }
 
 /**
- * shell_global_get_modifier_keys:
+ * shell_global_get_pointer:
  * @global: the #ShellGlobal
+ * @x: (out): the X coordinate of the pointer, in global coordinates
+ * @y: (out): the Y coordinate of the pointer, in global coordinates
+ * @mods: (out): the current set of modifier keys that are pressed down
  *
- * Gets the current set of modifier keys that are pressed down;
- * this is a wrapper around gdk_display_get_pointer() that strips
+ * Gets the pointer coordinates and current modifier key state.
+ * This is a wrapper around gdk_display_get_pointer() that strips
  * out any un-declared modifier flags, to make gjs happy; see
  * https://bugzilla.gnome.org/show_bug.cgi?id=597292.
- *
- * Return value: the current modifiers
  */
-GdkModifierType
-shell_global_get_modifier_keys (ShellGlobal *global)
+void
+shell_global_get_pointer (ShellGlobal         *global,
+                          int                 *x,
+                          int                 *y,
+                          ClutterModifierType *mods)
 {
-  GdkModifierType mods;
+  GdkModifierType raw_mods;
 
-  gdk_display_get_pointer (gdk_display_get_default (), NULL, NULL, NULL, &mods);
-  return mods & GDK_MODIFIER_MASK;
+  gdk_display_get_pointer (gdk_display_get_default (), NULL, x, y, &raw_mods);
+  *mods = raw_mods & GDK_MODIFIER_MASK;
 }
 
 /**
@@ -1017,3 +1296,336 @@ shell_get_event_state (ClutterEvent *event)
   ClutterModifierType state = clutter_event_get_state (event);
   return state & CLUTTER_MODIFIER_MASK;
 }
+
+/**
+ * shell_global_get_settings:
+ * @global: A #ShellGlobal
+ *
+ * Get the global GSettings instance.
+ *
+ * Return value: (transfer none) The GSettings object (transfer none).
+ */
+GSettings *
+shell_global_get_settings (ShellGlobal *global)
+{
+  return global->settings;
+}
+
+static void
+shell_popup_menu_position_func (GtkMenu   *menu,
+                                int       *x,
+                                int       *y,
+                                gboolean  *push_in,
+                                gpointer   user_data)
+{
+  *x = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (menu), "shell-menu-x"));
+  *y = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (menu), "shell-menu-y"));
+}
+
+/**
+ * shell_popup_menu:
+ * @menu: a #GtkMenu
+ * @button: mouse button that triggered the menu
+ * @time: timestamp of event that triggered the menu
+ * @menu_x: x coordinate to display the menu at
+ * @menu_y: y coordinate to display the menu at
+ *
+ * Wraps gtk_menu_popup(), but using @menu_x, @menu_y for the location
+ * rather than needing a callback.
+ **/
+void
+shell_popup_menu (GtkMenu *menu, int button, guint32 time,
+                  int menu_x, int menu_y)
+{
+  g_object_set_data (G_OBJECT (menu), "shell-menu-x", GINT_TO_POINTER (menu_x));
+  g_object_set_data (G_OBJECT (menu), "shell-menu-y", GINT_TO_POINTER (menu_y));
+
+  gtk_menu_popup (menu, NULL, NULL, shell_popup_menu_position_func, NULL,
+                  button, time);
+}
+
+/**
+ * shell_write_string_to_stream:
+ * @stream: a #GOutputStream
+ * @str: a UTF-8 string to write to @stream
+ * @error: location to store GError
+ *
+ * Write a string to a GOutputStream as UTF-8. This is a workaround
+ * for not having binary buffers in GJS.
+ *
+ * Return value: %TRUE if write succeeded
+ */
+gboolean
+shell_write_string_to_stream (GOutputStream *stream,
+                              const char    *str,
+                              GError       **error)
+{
+  return g_output_stream_write_all (stream, str, strlen (str),
+                                    NULL, NULL, error);
+}
+
+/**
+ * shell_global_get_current_time:
+ * @global: A #ShellGlobal
+ *
+ * Returns: the current X server time from the current Clutter, Gdk, or X
+ * event. If called from outside an event handler, this may return
+ * %Clutter.CURRENT_TIME (aka 0), or it may return a slightly
+ * out-of-date timestamp.
+ */
+guint32
+shell_global_get_current_time (ShellGlobal *global)
+{
+  guint32 time;
+  MetaDisplay *display;
+
+  /* meta_display_get_current_time() will return the correct time
+     when handling an X or Gdk event, but will return CurrentTime
+     from some Clutter event callbacks.
+
+     clutter_get_current_event_time() will return the correct time
+     from a Clutter event callback, but may return an out-of-date
+     timestamp if called at other times.
+
+     So we try meta_display_get_current_time() first, since we
+     can recognize a "wrong" answer from that, and then fall back
+     to clutter_get_current_event_time().
+   */
+
+  display = meta_screen_get_display (shell_global_get_screen (global));
+  time = meta_display_get_current_time (display);
+  if (time != CLUTTER_CURRENT_TIME)
+      return time;
+
+  return clutter_get_current_event_time ();
+}
+
+/**
+ * shell_global_get_app_launch_context:
+ * @global: A #ShellGlobal
+ *
+ * Create a #GAppLaunchContext set up with the correct timestamp, and
+ * targeted to activate on the current workspace.
+ *
+ * Return value: A new #GAppLaunchContext
+ */
+GAppLaunchContext *
+shell_global_create_app_launch_context (ShellGlobal *global)
+{
+  GdkAppLaunchContext *context;
+
+  context = gdk_app_launch_context_new ();
+  gdk_app_launch_context_set_timestamp (context, shell_global_get_current_time (global));
+
+  // Make sure that the app is opened on the current workspace even if
+  // the user switches before it starts
+  gdk_app_launch_context_set_desktop (context, meta_screen_get_active_workspace_index (shell_global_get_screen (global)));
+
+  return (GAppLaunchContext *)context;
+}
+
+/**
+ * shell_global_set_property_mutable:
+ * @global: the #ShellGlobal
+ * @object: the "path" to a JS object, starting from the root object.
+ *  (Eg, "global.stage" or "imports.gi.Gtk.Window.prototype")
+ * @property: a property on @object
+ * @mutable: %TRUE or %FALSE
+ *
+ * If @mutable is %TRUE, this clears the "permanent" and "readonly" flags
+ * on @property of @object. If @mutable is %FALSE, it sets them.
+ *
+ * You can use this to make it possible to modify properties that
+ * would otherwise be read-only from JavaScript.
+ *
+ * Return value: success or failure.
+ */
+gboolean
+shell_global_set_property_mutable (ShellGlobal *global,
+                                   const char  *object,
+                                   const char  *property,
+                                   gboolean     mutable)
+{
+  JSContext *context = gjs_context_get_native_context (global->js_context);
+  char **parts;
+  JSObject *obj;
+  jsval val = JSVAL_VOID;
+  int i;
+  jsuint attrs;
+  JSBool found;
+
+  JS_AddRoot (context, &val);
+
+  parts = g_strsplit (object, ".", -1);
+  obj = JS_GetGlobalObject (context);
+  for (i = 0; parts[i]; i++)
+    {
+      if (!JS_GetProperty (context, obj, parts[i], &val))
+        {
+          g_strfreev (parts);
+          JS_RemoveRoot (context, &val);
+          gjs_log_exception (context, NULL);
+          return FALSE;
+        }
+      obj = JSVAL_TO_OBJECT (val);
+    }
+  g_strfreev (parts);
+
+  if (!JS_GetPropertyAttributes (context, obj, property, &attrs, &found) || !found)
+    {
+      JS_RemoveRoot (context, &val);
+      gjs_log_exception (context, NULL);
+      return FALSE;
+    }
+
+  if (mutable)
+    attrs &= ~(JSPROP_PERMANENT | JSPROP_READONLY);
+  else
+    attrs |= (JSPROP_PERMANENT | JSPROP_READONLY);
+
+  JS_SetPropertyAttributes (context, obj, property, attrs, &found);
+
+  JS_RemoveRoot (context, &val);
+  return !gjs_log_exception (context, NULL);
+}
+
+typedef struct
+{
+  ShellLeisureFunction func;
+  gpointer user_data;
+  GDestroyNotify notify;
+} LeisureClosure;
+
+static gboolean
+run_leisure_functions (gpointer data)
+{
+  ShellGlobal *global = data;
+  GSList *closures;
+  GSList *iter;
+
+  global->leisure_function_id = 0;
+
+  /* We started more work since we scheduled the idle */
+  if (global->work_count > 0)
+    return FALSE;
+
+  /*
+   * We do call MAYBE_GC() here to free up some memory and
+   * prevent the GC from running when we are busy doing other things.
+   */
+  shell_global_maybe_gc (global);
+
+  /* No leisure closures, so we are done */
+  if (global->leisure_closures == NULL)
+    return FALSE;
+
+  closures = global->leisure_closures;
+  global->leisure_closures = NULL;
+
+  for (iter = closures; iter; iter = iter->next)
+    {
+      LeisureClosure *closure = closures->data;
+      closure->func (closure->user_data);
+
+      if (closure->notify)
+        closure->notify (closure->user_data);
+
+      g_slice_free (LeisureClosure, closure);
+    }
+
+  g_slist_free (closures);
+
+  return FALSE;
+}
+
+static void
+schedule_leisure_functions (ShellGlobal *global)
+{
+  /* This is called when we think we are ready to run leisure functions
+   * by our own accounting. We try to handle other types of business
+   * (like ClutterAnimation) by adding a low priority idle function.
+   *
+   * This won't work properly if the mainloop goes idle waiting for
+   * the vertical blanking interval or waiting for work being done
+   * in another thread.
+   */
+  if (!global->leisure_function_id)
+    global->leisure_function_id = g_idle_add_full (G_PRIORITY_LOW,
+                                                   run_leisure_functions,
+                                                   global, NULL);
+}
+
+/**
+ * shell_global_begin_work:
+ * @global: the #ShellGlobal
+ *
+ * Marks that we are currently doing work. This is used to to track
+ * whether we are busy for the purposes of shell_global_run_at_leisure().
+ * A count is kept and shell_global_end_work() must be called exactly
+ * as many times as shell_global_begin_work().
+ */
+void
+shell_global_begin_work (ShellGlobal *global)
+{
+  global->work_count++;
+}
+
+/**
+ * shell_global_end_work:
+ * @global: the #ShellGlobal
+ *
+ * Marks the end of work that we started with shell_global_begin_work().
+ * If no other work is ongoing and functions have been added with
+ * shell_global_run_at_leisure(), they will be run at the next
+ * opportunity.
+ */
+void
+shell_global_end_work (ShellGlobal *global)
+{
+  g_return_if_fail (global->work_count > 0);
+
+  global->work_count--;
+  if (global->work_count == 0)
+    schedule_leisure_functions (global);
+
+}
+
+/**
+ * shell_global_run_at_leisure:
+ * @global: the #ShellGlobal
+ * @func: function to call at leisure
+ * @user_data: data to pass to @func
+ * @notify: function to call to free @user_data
+ *
+ * Schedules a function to be called the next time the shell is idle.
+ * Idle means here no animations, no redrawing, and no ongoing background
+ * work. Since there is currently no way to hook into the Clutter master
+ * clock and know when is running, the implementation here is somewhat
+ * approximation. Animations done through the shell's Tweener module will
+ * be handled properly, but other animations may be detected as terminating
+ * early if they can be drawn fast enough so that the event loop goes idle
+ * between frames.
+ *
+ * The intent of this function is for performance measurement runs
+ * where a number of actions should be run serially and each action is
+ * timed individually. Using this function for other purposes will
+ * interfere with the ability to use it for performance measurement so
+ * should be avoided.
+ */
+void
+shell_global_run_at_leisure (ShellGlobal         *global,
+                             ShellLeisureFunction func,
+                             gpointer             user_data,
+                             GDestroyNotify       notify)
+{
+  LeisureClosure *closure = g_slice_new (LeisureClosure);
+  closure->func = func;
+  closure->user_data = user_data;
+  closure->notify = notify;
+
+  global->leisure_closures = g_slist_append (global->leisure_closures,
+                                             closure);
+
+  if (global->work_count == 0)
+    schedule_leisure_functions (global);
+}
