@@ -38,6 +38,10 @@
  *  <listitem>
  *   <para>indeterminate: the widget is showing the hint text</para>
  *  </listitem>
+ *  <listitem>
+ *   <para>hover: the widget is showing the hint text and is underneath the
+ *                pointer</para>
+ *  </listitem>
  * </itemizedlist>
  */
 
@@ -61,6 +65,7 @@
 #include "st-texture-cache.h"
 #include "st-marshal.h"
 #include "st-clipboard.h"
+#include "st-private.h"
 
 #define HAS_FOCUS(actor) (clutter_actor_get_stage (actor) && clutter_stage_get_key_focus ((ClutterStage *) clutter_actor_get_stage (actor)) == actor)
 
@@ -97,6 +102,8 @@ struct _StEntryPrivate
   ClutterActor *secondary_icon;
 
   gfloat        spacing;
+
+  gboolean      hint_visible;
 };
 
 static guint entry_signals[LAST_SIGNAL] = { 0, };
@@ -165,6 +172,8 @@ st_entry_dispose (GObject *object)
       clutter_actor_destroy (priv->entry);
       priv->entry = NULL;
     }
+
+  G_OBJECT_CLASS (st_entry_parent_class)->dispose (object);
 }
 
 static void
@@ -174,6 +183,8 @@ st_entry_finalize (GObject *object)
 
   g_free (priv->hint);
   priv->hint = NULL;
+
+  G_OBJECT_CLASS (st_entry_parent_class)->finalize (object);
 }
 
 static void
@@ -184,12 +195,16 @@ st_entry_style_changed (StWidget *self)
   ClutterColor color;
   const PangoFontDescription *font;
   gchar *font_string;
+  gdouble size;
 
   theme_node = st_widget_get_theme_node (self);
  
   st_theme_node_get_foreground_color (theme_node, &color);
   clutter_text_set_color (CLUTTER_TEXT (priv->entry), &color);
 
+  if (st_theme_node_get_length (theme_node, "caret-size", FALSE, &size))
+    clutter_text_set_cursor_size (CLUTTER_TEXT (priv->entry), (int)(.5 + size));
+
   if (st_theme_node_get_color (theme_node, "caret-color", FALSE, &color))
     clutter_text_set_cursor_color (CLUTTER_TEXT (priv->entry), &color);
 
@@ -371,12 +386,14 @@ clutter_text_focus_in_cb (ClutterText  *text,
   StEntryPrivate *priv = ST_ENTRY_PRIV (actor);
 
   /* remove the hint if visible */
-  if (priv->hint
-      && !strcmp (clutter_text_get_text (text), priv->hint))
+  if (priv->hint && priv->hint_visible)
     {
+      priv->hint_visible = FALSE;
+
       clutter_text_set_text (text, "");
     }
-  st_widget_set_style_pseudo_class (ST_WIDGET (actor), "focus");
+  st_widget_remove_style_pseudo_class (ST_WIDGET (actor), "indeterminate");
+  st_widget_add_style_pseudo_class (ST_WIDGET (actor), "focus");
   clutter_text_set_cursor_visible (text, TRUE);
 }
 
@@ -386,15 +403,15 @@ clutter_text_focus_out_cb (ClutterText  *text,
 {
   StEntryPrivate *priv = ST_ENTRY_PRIV (actor);
 
+  st_widget_remove_style_pseudo_class (ST_WIDGET (actor), "focus");
+
   /* add a hint if the entry is empty */
   if (priv->hint && !strcmp (clutter_text_get_text (text), ""))
     {
+      priv->hint_visible = TRUE;
+
       clutter_text_set_text (text, priv->hint);
-      st_widget_set_style_pseudo_class (ST_WIDGET (actor), "indeterminate");
-    }
-  else
-    {
-      st_widget_set_style_pseudo_class (ST_WIDGET (actor), NULL);
+      st_widget_add_style_pseudo_class (ST_WIDGET (actor), "indeterminate");
     }
   clutter_text_set_cursor_visible (text, FALSE);
 }
@@ -561,6 +578,31 @@ st_entry_key_focus_in (ClutterActor *actor)
   clutter_actor_grab_key_focus (priv->entry);
 }
 
+static gboolean
+st_entry_enter_event (ClutterActor         *actor,
+                      ClutterCrossingEvent *event)
+{
+  StEntryPrivate *priv = ST_ENTRY_PRIV (actor);
+
+  if (_st_actor_contains (actor, event->source)
+      && priv->hint && priv->hint_visible)
+    {
+      st_widget_set_hover (ST_WIDGET (actor), TRUE);
+    }
+
+  return CLUTTER_ACTOR_CLASS (st_entry_parent_class)->enter_event (actor, event);
+}
+
+static gboolean
+st_entry_leave_event (ClutterActor         *actor,
+                      ClutterCrossingEvent *event)
+{
+  if (!_st_actor_contains (actor, event->related))
+    st_widget_set_hover (ST_WIDGET (actor), FALSE);
+
+  return CLUTTER_ACTOR_CLASS (st_entry_parent_class)->leave_event (actor, event);
+}
+
 static void
 st_entry_class_init (StEntryClass *klass)
 {
@@ -586,6 +628,8 @@ st_entry_class_init (StEntryClass *klass)
 
   actor_class->key_press_event = st_entry_key_press_event;
   actor_class->key_focus_in = st_entry_key_focus_in;
+  actor_class->enter_event = st_entry_enter_event;
+  actor_class->leave_event = st_entry_leave_event;
 
   widget_class->style_changed = st_entry_style_changed;
 
@@ -701,13 +745,16 @@ st_entry_get_text (StEntry *entry)
 {
   g_return_val_if_fail (ST_IS_ENTRY (entry), NULL);
 
-  return clutter_text_get_text (CLUTTER_TEXT (entry->priv->entry));
+  if (entry->priv->hint_visible)
+    return "";
+  else
+    return clutter_text_get_text (CLUTTER_TEXT (entry->priv->entry));
 }
 
 /**
  * st_entry_set_text:
  * @entry: a #StEntry
- * @text: text to set the entry to
+ * @text: (allow-none): text to set the entry to
  *
  * Sets the text displayed on the entry
  */
@@ -727,14 +774,14 @@ st_entry_set_text (StEntry     *entry,
       && !HAS_FOCUS (priv->entry))
     {
       text = priv->hint;
-      st_widget_set_style_pseudo_class (ST_WIDGET (entry), "indeterminate");
+      priv->hint_visible = TRUE;
+      st_widget_add_style_pseudo_class (ST_WIDGET (entry), "indeterminate");
     }
   else
     {
-      if (HAS_FOCUS (priv->entry))
-        st_widget_set_style_pseudo_class (ST_WIDGET (entry), "focus");
-      else
-        st_widget_set_style_pseudo_class (ST_WIDGET (entry), NULL);
+      st_widget_remove_style_pseudo_class (ST_WIDGET (entry), "indeterminate");
+
+      priv->hint_visible = FALSE;
     }
 
   clutter_text_set_text (CLUTTER_TEXT (priv->entry), text);
@@ -762,7 +809,7 @@ st_entry_get_clutter_text (StEntry *entry)
 /**
  * st_entry_set_hint_text:
  * @entry: a #StEntry
- * @text: text to set as the entry hint
+ * @text: (allow-none): text to set as the entry hint
  *
  * Sets the text to display when the entry is empty and unfocused. When the
  * entry is displaying the hint, it has a pseudo class of "indeterminate".
@@ -782,10 +829,13 @@ st_entry_set_hint_text (StEntry     *entry,
 
   priv->hint = g_strdup (text);
 
-  if (!strcmp (clutter_text_get_text (CLUTTER_TEXT (priv->entry)), ""))
+  if (!strcmp (clutter_text_get_text (CLUTTER_TEXT (priv->entry)), "")
+      && !HAS_FOCUS (priv->entry))
     {
+      priv->hint_visible = TRUE;
+
       clutter_text_set_text (CLUTTER_TEXT (priv->entry), priv->hint);
-      st_widget_set_style_pseudo_class (ST_WIDGET (entry), "indeterminate");
+      st_widget_add_style_pseudo_class (ST_WIDGET (entry), "indeterminate");
     }
 }
 
@@ -842,9 +892,7 @@ _st_entry_set_icon_from_file (StEntry       *entry,
 
       cache = st_texture_cache_get_default ();
 
-
-
-      *icon = (ClutterActor*) st_texture_cache_get_texture (cache, filename);
+      *icon = (ClutterActor*) st_texture_cache_load_file_simple (cache, filename);
 
       clutter_actor_set_reactive (*icon, TRUE);
       clutter_actor_set_parent (*icon, CLUTTER_ACTOR (entry));
@@ -858,7 +906,7 @@ _st_entry_set_icon_from_file (StEntry       *entry,
 /**
  * st_entry_set_primary_icon_from_file:
  * @entry: a #StEntry
- * @filename: filename of an icon
+ * @filename: (allow-none): filename of an icon
  *
  * Set the primary icon of the entry to the given filename
  */
@@ -879,7 +927,7 @@ st_entry_set_primary_icon_from_file (StEntry     *entry,
 /**
  * st_entry_set_secondary_icon_from_file:
  * @entry: a #StEntry
- * @filename: filename of an icon
+ * @filename: (allow-none): filename of an icon
  *
  * Set the primary icon of the entry to the given filename
  */
