@@ -5,78 +5,25 @@
 
 #include "st-theme-private.h"
 #include "st-theme-context.h"
-#include "st-theme-node.h"
+#include "st-theme-node-private.h"
 
 static void st_theme_node_init               (StThemeNode          *node);
 static void st_theme_node_class_init         (StThemeNodeClass     *klass);
+static void st_theme_node_dispose           (GObject                 *object);
 static void st_theme_node_finalize           (GObject                 *object);
 
-struct _StThemeNode {
-  GObject parent;
-
-  StThemeContext *context;
-  StThemeNode *parent_node;
-  StTheme *theme;
-
-  PangoFontDescription *font_desc;
-
-  ClutterColor background_color;
-  /* If gradient is set, then background_color is the gradient start */
-  StGradientType background_gradient_type;
-  ClutterColor background_gradient_end;
-
-  ClutterColor foreground_color;
-  ClutterColor border_color[4];
-
-  int border_width[4];
-  int border_radius[4];
-  guint padding[4];
-
-  int width;
-  int height;
-  int min_width;
-  int min_height;
-  int max_width;
-  int max_height;
-
-  char *background_image;
-  StBorderImage *border_image;
-  StShadow *shadow;
-
-  GType element_type;
-  char *element_id;
-  char *element_class;
-  char *pseudo_class;
-  char *inline_style;
-
-  CRDeclaration **properties;
-  int n_properties;
-
-  /* We hold onto these separately so we can destroy them on finalize */
-  CRDeclaration *inline_properties;
-
-  guint properties_computed : 1;
-  guint geometry_computed : 1;
-  guint background_computed : 1;
-  guint foreground_computed : 1;
-  guint border_image_computed : 1;
-  guint shadow_computed : 1;
-  guint link_type : 2;
-};
-
-struct _StThemeNodeClass {
-  GObjectClass parent_class;
-
-};
-
 static const ClutterColor BLACK_COLOR = { 0, 0, 0, 0xff };
 static const ClutterColor TRANSPARENT_COLOR = { 0, 0, 0, 0 };
 
+extern gfloat st_slow_down_factor;
+
 G_DEFINE_TYPE (StThemeNode, st_theme_node, G_TYPE_OBJECT)
 
 static void
 st_theme_node_init (StThemeNode *node)
 {
+  node->transition_duration = -1;
+  _st_theme_node_init_drawing_state (node);
 }
 
 static void
@@ -84,9 +31,43 @@ st_theme_node_class_init (StThemeNodeClass *klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
+  object_class->dispose = st_theme_node_dispose;
   object_class->finalize = st_theme_node_finalize;
 }
 
+
+static void
+st_theme_node_dispose (GObject *gobject)
+{
+  StThemeNode *node = ST_THEME_NODE (gobject);
+
+  if (node->context)
+    {
+      g_object_unref (node->context);
+      node->context = NULL;
+    }
+
+  if (node->theme)
+    {
+      g_object_unref (node->theme);
+      node->theme = NULL;
+    }
+
+  if (node->parent_node)
+    {
+      g_object_unref (node->parent_node);
+      node->parent_node = NULL;
+    }
+
+  if (node->border_image)
+    {
+      g_object_unref (node->border_image);
+      node->border_image = NULL;
+    }
+
+  G_OBJECT_CLASS (st_theme_node_parent_class)->dispose (gobject);
+}
+
 static void
 st_theme_node_finalize (GObject *object)
 {
@@ -116,12 +97,6 @@ st_theme_node_finalize (GObject *object)
       node->font_desc = NULL;
     }
 
-  if (node->border_image)
-    {
-      g_object_unref (node->border_image);
-      node->border_image = NULL;
-    }
-
   if (node->shadow)
     {
       st_shadow_free (node->shadow);
@@ -131,6 +106,8 @@ st_theme_node_finalize (GObject *object)
   if (node->background_image)
     g_free (node->background_image);
 
+  _st_theme_node_free_drawing_state (node);
+
   G_OBJECT_CLASS (st_theme_node_parent_class)->finalize (object);
 }
 
@@ -259,6 +236,49 @@ st_theme_node_get_pseudo_class (StThemeNode *node)
   return node->pseudo_class;
 }
 
+/**
+ * st_theme_node_equal:
+ * @node_a: first #StThemeNode
+ * @node_b: second #StThemeNode
+ *
+ * Compare two #StThemeNodes. Two nodes which compare equal will match
+ * the same CSS rules and have the same style properties. However, two
+ * nodes that have ended up with identical style properties do not
+ * necessarily compare equal.
+ * In detail, @node_a and @node_b are considered equal iff
+ * <itemizedlist>
+ *   <listitem>
+ *     <para>they share the same #StTheme and #StThemeContext</para>
+ *   </listitem>
+ *   <listitem>
+ *     <para>they have the same parent</para>
+ *   </listitem>
+ *   <listitem>
+ *     <para>they have the same element type</para>
+ *   </listitem>
+ *   <listitem>
+ *     <para>their id, class, pseudo-class and inline-style match</para>
+ *   </listitem>
+ * </itemizedlist>
+ *
+ * Returns: %TRUE if @node_a equals @node_b
+ */
+gboolean
+st_theme_node_equal (StThemeNode *node_a, StThemeNode *node_b)
+{
+  g_return_val_if_fail (ST_IS_THEME_NODE (node_a), FALSE);
+  g_return_val_if_fail (ST_IS_THEME_NODE (node_b), FALSE);
+
+  return node_a->parent_node == node_b->parent_node &&
+         node_a->context == node_b->context &&
+         node_a->theme == node_b->theme &&
+         node_a->element_type == node_b->element_type &&
+         !g_strcmp0 (node_a->element_id, node_b->element_id) &&
+         !g_strcmp0 (node_a->element_class, node_b->element_class) &&
+         !g_strcmp0 (node_a->pseudo_class, node_b->pseudo_class) &&
+         !g_strcmp0 (node_a->inline_style, node_b->inline_style);
+}
+
 static void
 ensure_properties (StThemeNode *node)
 {
@@ -925,7 +945,7 @@ do_border_property (StThemeNode   *node,
 
   if (strcmp (property_name, "") == 0)
     {
-      /* Set value for width/color/node in any order */
+      /* Set value for width/color/style in any order */
       CRTerm *term;
 
       for (term = decl->value; term; term = term->next)
@@ -947,7 +967,6 @@ do_border_property (StThemeNode   *node,
                 }
               else if (strcmp (ident, "dotted") == 0 ||
                        strcmp (ident, "dashed") == 0 ||
-                       strcmp (ident, "solid") == 0 ||
                        strcmp (ident, "double") == 0 ||
                        strcmp (ident, "groove") == 0 ||
                        strcmp (ident, "ridge") == 0 ||
@@ -1018,6 +1037,97 @@ do_border_property (StThemeNode   *node,
     }
 }
 
+static void
+do_outline_property (StThemeNode   *node,
+                     CRDeclaration *decl)
+{
+  const char *property_name = decl->property->stryng->str + 7; /* Skip 'outline' */
+  ClutterColor color;
+  gboolean color_set = FALSE;
+  int width;
+  gboolean width_set = FALSE;
+
+  if (strcmp (property_name, "") == 0)
+    {
+      /* Set value for width/color/style in any order */
+      CRTerm *term;
+
+      for (term = decl->value; term; term = term->next)
+        {
+          GetFromTermResult result;
+
+          if (term->type == TERM_IDENT)
+            {
+              const char *ident = term->content.str->stryng->str;
+              if (strcmp (ident, "none") == 0 || strcmp (ident, "hidden") == 0)
+                {
+                  width = 0.;
+                  continue;
+                }
+              else if (strcmp (ident, "solid") == 0)
+                {
+                  /* The only thing we support */
+                  continue;
+                }
+              else if (strcmp (ident, "dotted") == 0 ||
+                       strcmp (ident, "dashed") == 0 ||
+                       strcmp (ident, "double") == 0 ||
+                       strcmp (ident, "groove") == 0 ||
+                       strcmp (ident, "ridge") == 0 ||
+                       strcmp (ident, "inset") == 0 ||
+                       strcmp (ident, "outset") == 0)
+                {
+                  /* Treat the same as solid */
+                  continue;
+                }
+
+              /* Presumably a color, fall through */
+            }
+
+          if (term->type == TERM_NUMBER)
+            {
+              result = get_length_from_term_int (node, term, FALSE, &width);
+              if (result != VALUE_NOT_FOUND)
+                {
+                  width_set = result == VALUE_FOUND;
+                  continue;
+                }
+            }
+
+          result = get_color_from_term (node, term, &color);
+          if (result != VALUE_NOT_FOUND)
+            {
+              color_set = result == VALUE_FOUND;
+              continue;
+            }
+        }
+
+    }
+  else if (strcmp (property_name, "-color") == 0)
+    {
+      if (decl->value == NULL || decl->value->next != NULL)
+        return;
+
+      if (get_color_from_term (node, decl->value, &color) == VALUE_FOUND)
+        /* Ignore inherit */
+        color_set = TRUE;
+    }
+  else if (strcmp (property_name, "-width") == 0)
+    {
+      if (decl->value == NULL || decl->value->next != NULL)
+        return;
+
+      if (get_length_from_term_int (node, decl->value, FALSE, &width) == VALUE_FOUND)
+        /* Ignore inherit */
+        width_set = TRUE;
+    }
+
+  if (color_set)
+    node->outline_color = color;
+  if (width_set)
+    node->outline_width = width;
+}
+
 static void
 do_padding_property_term (StThemeNode *node,
                           CRTerm      *term,
@@ -1108,8 +1218,8 @@ do_size_property (StThemeNode   *node,
   get_length_from_term_int (node, decl->value, FALSE, node_value);
 }
 
-static void
-ensure_geometry (StThemeNode *node)
+void
+_st_theme_node_ensure_geometry (StThemeNode *node)
 {
   int i, j;
 
@@ -1126,6 +1236,9 @@ ensure_geometry (StThemeNode *node)
       node->border_color[j] = TRANSPARENT_COLOR;
     }
 
+  node->outline_width = 0;
+  node->outline_color = TRANSPARENT_COLOR;
+
   node->width = -1;
   node->height = -1;
   node->min_width = -1;
@@ -1140,6 +1253,8 @@ ensure_geometry (StThemeNode *node)
 
       if (g_str_has_prefix (property_name, "border"))
         do_border_property (node, decl);
+      else if (g_str_has_prefix (property_name, "outline"))
+        do_outline_property (node, decl);
       else if (g_str_has_prefix (property_name, "padding"))
         do_padding_property (node, decl);
       else if (strcmp (property_name, "width") == 0)
@@ -1167,8 +1282,6 @@ ensure_geometry (StThemeNode *node)
       else if (node->width > node->max_width)
         node->width = node->max_width;
     }
-  else
-    node->width = node->min_width;
 
   if (node->height != -1)
     {
@@ -1181,8 +1294,6 @@ ensure_geometry (StThemeNode *node)
       else if (node->height > node->max_height)
         node->height = node->max_height;
     }
-  else
-    node->height = node->min_height;
 }
 
 int
@@ -1192,7 +1303,7 @@ st_theme_node_get_border_width (StThemeNode *node,
   g_return_val_if_fail (ST_IS_THEME_NODE (node), 0.);
   g_return_val_if_fail (side >= ST_SIDE_TOP && side <= ST_SIDE_LEFT, 0.);
 
-  ensure_geometry (node);
+  _st_theme_node_ensure_geometry (node);
 
   return node->border_width[side];
 }
@@ -1204,17 +1315,38 @@ st_theme_node_get_border_radius (StThemeNode *node,
   g_return_val_if_fail (ST_IS_THEME_NODE (node), 0.);
   g_return_val_if_fail (corner >= ST_CORNER_TOPLEFT && corner <= ST_CORNER_BOTTOMLEFT, 0.);
 
-  ensure_geometry (node);
+  _st_theme_node_ensure_geometry (node);
 
   return node->border_radius[corner];
 }
 
+int
+st_theme_node_get_outline_width (StThemeNode  *node)
+{
+  g_return_val_if_fail (ST_IS_THEME_NODE (node), 0);
+
+  _st_theme_node_ensure_geometry (node);
+
+  return node->outline_width;
+}
+
+void
+st_theme_node_get_outline_color (StThemeNode  *node,
+                                 ClutterColor *color)
+{
+  g_return_if_fail (ST_IS_THEME_NODE (node));
+
+  _st_theme_node_ensure_geometry (node);
+
+  *color = node->outline_color;
+}
+
 int
 st_theme_node_get_width (StThemeNode *node)
 {
   g_return_val_if_fail (ST_IS_THEME_NODE (node), -1);
 
-  ensure_geometry (node);
+  _st_theme_node_ensure_geometry (node);
   return node->width;
 }
 
@@ -1223,7 +1355,7 @@ st_theme_node_get_height (StThemeNode *node)
 {
   g_return_val_if_fail (ST_IS_THEME_NODE (node), -1);
 
-  ensure_geometry (node);
+  _st_theme_node_ensure_geometry (node);
   return node->height;
 }
 
@@ -1232,7 +1364,7 @@ st_theme_node_get_min_width (StThemeNode *node)
 {
   g_return_val_if_fail (ST_IS_THEME_NODE (node), -1);
 
-  ensure_geometry (node);
+  _st_theme_node_ensure_geometry (node);
   return node->min_width;
 }
 
@@ -1241,7 +1373,7 @@ st_theme_node_get_min_height (StThemeNode *node)
 {
   g_return_val_if_fail (ST_IS_THEME_NODE (node), -1);
 
-  ensure_geometry (node);
+  _st_theme_node_ensure_geometry (node);
   return node->min_height;
 }
 
@@ -1250,7 +1382,7 @@ st_theme_node_get_max_width (StThemeNode *node)
 {
   g_return_val_if_fail (ST_IS_THEME_NODE (node), -1);
 
-  ensure_geometry (node);
+  _st_theme_node_ensure_geometry (node);
   return node->max_width;
 }
 
@@ -1259,7 +1391,7 @@ st_theme_node_get_max_height (StThemeNode *node)
 {
   g_return_val_if_fail (ST_IS_THEME_NODE (node), -1);
 
-  ensure_geometry (node);
+  _st_theme_node_ensure_geometry (node);
   return node->max_height;
 }
 
@@ -1281,8 +1413,8 @@ get_background_color_from_term (StThemeNode  *node,
   return result;
 }
 
-static void
-ensure_background (StThemeNode *node)
+void
+_st_theme_node_ensure_background (StThemeNode *node)
 {
   int i;
 
@@ -1429,7 +1561,7 @@ st_theme_node_get_background_color (StThemeNode  *node,
 {
   g_return_if_fail (ST_IS_THEME_NODE (node));
 
-  ensure_background (node);
+  _st_theme_node_ensure_background (node);
 
   *color = node->background_color;
 }
@@ -1439,7 +1571,7 @@ st_theme_node_get_background_image (StThemeNode *node)
 {
   g_return_val_if_fail (ST_IS_THEME_NODE (node), NULL);
 
-  ensure_background (node);
+  _st_theme_node_ensure_background (node);
 
   return node->background_image;
 }
@@ -1500,7 +1632,7 @@ st_theme_node_get_background_gradient (StThemeNode    *node,
 {
   g_return_if_fail (ST_IS_THEME_NODE (node));
 
-  ensure_background (node);
+  _st_theme_node_ensure_background (node);
 
   *type = node->background_gradient_type;
   if (*type != ST_GRADIENT_NONE)
@@ -1518,7 +1650,7 @@ st_theme_node_get_border_color (StThemeNode  *node,
   g_return_if_fail (ST_IS_THEME_NODE (node));
   g_return_if_fail (side >= ST_SIDE_TOP && side <= ST_SIDE_LEFT);
 
-  ensure_geometry (node);
+  _st_theme_node_ensure_geometry (node);
 
   *color = node->border_color[side];
 }
@@ -1530,11 +1662,38 @@ st_theme_node_get_padding (StThemeNode *node,
   g_return_val_if_fail (ST_IS_THEME_NODE (node), 0.);
   g_return_val_if_fail (side >= ST_SIDE_TOP && side <= ST_SIDE_LEFT, 0.);
 
-  ensure_geometry (node);
+  _st_theme_node_ensure_geometry (node);
 
   return node->padding[side];
 }
 
+/**
+ * st_theme_node_get_transition_duration:
+ * @node: an #StThemeNode
+ *
+ * Get the value of the transition-duration property, which
+ * specifies the transition time between the previous #StThemeNode
+ * and @node.
+ *
+ * Returns: the node's transition duration in milliseconds
+ */
+int
+st_theme_node_get_transition_duration (StThemeNode *node)
+{
+  gdouble value = 0.0;
+
+  g_return_val_if_fail (ST_IS_THEME_NODE (node), 0);
+
+  if (node->transition_duration > -1)
+    return st_slow_down_factor * node->transition_duration;
+
+  st_theme_node_get_double (node, "transition-duration", FALSE, &value);
+
+  node->transition_duration = (int)value;
+
+  return st_slow_down_factor * node->transition_duration;
+}
+
 StTextDecoration
 st_theme_node_get_text_decoration (StThemeNode *node)
 {
@@ -1602,6 +1761,53 @@ st_theme_node_get_text_decoration (StThemeNode *node)
   return 0;
 }
 
+StTextAlign
+st_theme_node_get_text_align(StThemeNode *node)
+{
+  int i;
+
+  ensure_properties(node);
+
+  for (i = node->n_properties - 1; i >= 0; i--)
+    {
+      CRDeclaration *decl = node->properties[i];
+
+      if (strcmp(decl->property->stryng->str, "text-align") == 0)
+        {
+          CRTerm *term = decl->value;
+
+          if (term->type != TERM_IDENT || term->next)
+            continue;
+
+          if (strcmp(term->content.str->stryng->str, "inherit") == 0)
+            {
+              if (node->parent_node)
+                return st_theme_node_get_text_align(node->parent_node);
+              return ST_TEXT_ALIGN_LEFT;
+            }
+          else if (strcmp(term->content.str->stryng->str, "left") == 0)
+            {
+              return ST_TEXT_ALIGN_LEFT;
+            }
+          else if (strcmp(term->content.str->stryng->str, "right") == 0)
+            {
+              return ST_TEXT_ALIGN_RIGHT;
+            }
+          else if (strcmp(term->content.str->stryng->str, "center") == 0)
+            {
+              return ST_TEXT_ALIGN_CENTER;
+            }
+          else if (strcmp(term->content.str->stryng->str, "justify") == 0)
+            {
+              return ST_TEXT_ALIGN_JUSTIFY;
+            }
+        }
+    }
+  if(node->parent_node)
+    return st_theme_node_get_text_align(node->parent_node);
+  return ST_TEXT_ALIGN_LEFT;
+}
+
 static gboolean
 font_family_from_terms (CRTerm *term,
                         char  **family)
@@ -2246,6 +2452,7 @@ st_theme_node_get_shadow (StThemeNode *node)
           gdouble xoffset = 0.;
           gdouble yoffset = 0.;
           gdouble blur = 0.;
+          gdouble spread = 0.;
           int n_offsets = 0;
 
           for (term = decl->value; term; term = term->next)
@@ -2274,6 +2481,13 @@ st_theme_node_get_shadow (StThemeNode *node)
                               g_warning ("Negative blur values are "
                                          "not allowed");
                           blur = value;
+                          break;
+                        case 3:
+                          if (multiplier < 0)
+                              g_warning ("Negative spread values are "
+                                         "not allowed");
+                          spread = value;
+                          break;
                         }
                       continue;
                     }
@@ -2285,7 +2499,9 @@ st_theme_node_get_shadow (StThemeNode *node)
                   continue;
                 }
             }
-          node->shadow = st_shadow_new (&color, xoffset, yoffset, blur);
+          node->shadow = st_shadow_new (&color,
+                                        xoffset, yoffset,
+                                        blur, spread);
 
           return node->shadow;
         }
@@ -2353,7 +2569,7 @@ st_theme_node_adjust_preferred_width (StThemeNode  *node,
 
   g_return_if_fail (ST_IS_THEME_NODE (node));
 
-  ensure_geometry (node);
+  _st_theme_node_ensure_geometry (node);
 
   width_inc = get_width_inc (node);
 
@@ -2420,7 +2636,7 @@ st_theme_node_adjust_preferred_height (StThemeNode  *node,
 
   g_return_if_fail (ST_IS_THEME_NODE (node));
 
-  ensure_geometry (node);
+  _st_theme_node_ensure_geometry (node);
 
   height_inc = get_height_inc (node);
 
@@ -2461,7 +2677,7 @@ st_theme_node_get_content_box (StThemeNode           *node,
 
   g_return_if_fail (ST_IS_THEME_NODE (node));
 
-  ensure_geometry (node);
+  _st_theme_node_ensure_geometry (node);
 
   avail_width = allocation->x2 - allocation->x1;
   avail_height = allocation->y2 - allocation->y1;
@@ -2485,6 +2701,49 @@ st_theme_node_get_content_box (StThemeNode           *node,
   content_box->y2 = (int)(0.5 + content_box->y1 + content_height);
 }
 
+/**
+ * st_theme_node_get_paint_box:
+ * @node: a #StThemeNode
+ * @allocation: the box allocated to a #ClutterActor
+ * @paint_box: computed box occupied when painting the actor
+ *
+ * Gets the box used to paint the actor, including the area occupied
+ * by properties which paint outside the actor's assigned allocation.
+ * When painting @node to an offscreen buffer, this function can be
+ * used to determine the necessary size of the buffer.
+ */
+void
+st_theme_node_get_paint_box (StThemeNode           *node,
+                             const ClutterActorBox *actor_box,
+                             ClutterActorBox       *paint_box)
+{
+  StShadow *shadow;
+  ClutterActorBox shadow_box;
+  int outline_width;
+
+  g_return_if_fail (ST_IS_THEME_NODE (node));
+  g_return_if_fail (actor_box != NULL);
+  g_return_if_fail (paint_box != NULL);
+
+  shadow = st_theme_node_get_shadow (node);
+  outline_width = st_theme_node_get_outline_width (node);
+  if (!shadow && !outline_width)
+    {
+      *paint_box = *actor_box;
+      return;
+    }
+
+  if (shadow)
+    st_shadow_get_box (shadow, actor_box, &shadow_box);
+  else
+    shadow_box = *actor_box;
+
+  paint_box->x1 = MIN (actor_box->x1 - outline_width, shadow_box.x1);
+  paint_box->x2 = MAX (actor_box->x2 + outline_width, shadow_box.x2);
+  paint_box->y1 = MIN (actor_box->y1 - outline_width, shadow_box.y1);
+  paint_box->y2 = MAX (actor_box->y2 + outline_width, shadow_box.y2);
+}
+
 
 /**
  * st_theme_node_geometry_equal:
@@ -2501,8 +2760,8 @@ st_theme_node_geometry_equal (StThemeNode *node,
 {
   StSide side;
 
-  ensure_geometry (node);
-  ensure_geometry (other);
+  _st_theme_node_ensure_geometry (node);
+  _st_theme_node_ensure_geometry (other);
 
   for (side = ST_SIDE_TOP; side <= ST_SIDE_LEFT; side++)
     {
